"""PyPI integration."""

from __future__ import annotations

import subprocess
from pathlib import Path
from tempfile import TemporaryDirectory
from textwrap import dedent
from typing import Annotated as An

from build import ProjectBuilder
from build.env import DefaultIsolatedEnv
from failprint import Capture
from pypi_insiders.logger import log_captured, logger, redirect_output_to_logging
from twine.commands.upload import upload
from twine.settings import Settings
from typing_extensions import Doc

# TODO: Rewrite as a proper Client subclass.


def _git_user_name(default: str = "") -> str:
    return subprocess.getoutput("git config user.name").strip() or default  # noqa: S605,S607


def _git_user_email(default: str = "") -> str:
    return subprocess.getoutput("git config user.email").strip() or default  # noqa: S605,S607


def reserve_pypi(
    username: An[str, Doc("Username on PyPI.")],
    name: An[str, Doc("Name to reserve.")],
    description: An[str, Doc("Description of the project on PyPI.")],
) -> None:
    """Reserve a name on PyPI."""
    with TemporaryDirectory(prefix="insiders-") as tmpdir:
        repo = Path(tmpdir, name)
        repo.mkdir()
        dist = repo / "dist"

        logger.info(f"Preparing package {name}, {description}")
        repo.joinpath("pyproject.toml").write_text(
            dedent(
                f"""
                [project]
                name = "{name}"
                version = "0.0.0"
                description = "{description} Available to sponsors only."
                authors = [{{name = "{_git_user_name()}", email = "{_git_user_email()}"}}]
                readme = "README.md"
                requires-python = ">=3.8"
                classifiers = ["Development Status :: 1 - Planning"]
                """,
            ).lstrip(),
        )

        repo.joinpath("README.md").write_text(
            dedent(
                f"""
                # {name}

                {description}

                This project is currently available to [sponsors](https://github.com/sponsors/{username}) only.
                See https://{username}.github.io/{name}/insiders.
                """,
            ).lstrip(),
        )

        logger.info("Building distributions")
        for distribution in ("sdist", "wheel"):
            with DefaultIsolatedEnv() as env:
                builder = ProjectBuilder.from_isolated_env(env, repo)
                env.install(builder.build_system_requires)
                with Capture.BOTH.here() as captured:
                    env.install(builder.get_requires_for_build(distribution))
                    builder.build(distribution, str(dist))
                log_captured(str(captured), level="debug", pkg="build")

        logger.info("Uploading distributions")
        with redirect_output_to_logging(stdout_level="debug"):
            upload(
                Settings(
                    non_interactive=True,
                    skip_existing=True,
                    username="__token__",
                    disable_progress_bar=True,
                    verbose=True,
                ),
                [str(file) for file in dist.iterdir()],
            )



"""Logging utilities."""

from __future__ import annotations

import logging
import subprocess
import sys
import time
from contextlib import closing, contextmanager, redirect_stderr, redirect_stdout
from io import StringIO
from typing import TYPE_CHECKING, Annotated

from loguru import logger
from typing_extensions import Doc

if TYPE_CHECKING:
    from collections.abc import Callable, Iterator
    from pathlib import Path
    from typing import Any

    from loguru import Record


def double_brackets(message: str) -> str:
    """Double `{` and `}` in log messages to prevent formatting errors.

    Parameters:
        message: The message to transform.

    Returns:
        The updated message.
    """
    return message.replace("{", "{{").replace("}", "}}")


def run(*args: str | Path, **kwargs: Any) -> str:
    """Run a subprocess, log its standard output and error, return its output.

    Parameters:
        *args: Command line arguments.
        **kwargs: Additional arguments passed to [subprocess.Popen][].

    Returns:
        The process standard output.
    """
    args_str = double_brackets(str(args))
    kwargs_str = double_brackets(str(kwargs))
    logger.debug(f"Running subprocess with args={args_str}, kwargs={kwargs_str}")
    process = subprocess.Popen(  # noqa: S603
        args,
        **kwargs,
        text=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    stdout = []
    while True:
        stdout_line = process.stdout.readline().strip()  # type: ignore[union-attr]
        stderr_line = process.stderr.readline().strip()  # type: ignore[union-attr]
        if stdout_line:
            logger.debug(f"STDOUT: {double_brackets(stdout_line)}", pkg=args[0])
            stdout.append(stdout_line)
        if stderr_line:
            logger.debug(f"STDERR: {double_brackets(stderr_line)}", pkg=args[0])
        if not stdout_line and not stderr_line:
            break
    process.wait()
    return "\n".join(stdout)


class _TextBuffer(StringIO):
    class _BytesBuffer:
        def __init__(self, text_buffer: _TextBuffer) -> None:
            self._text_buffer = text_buffer

        def flush(self) -> None: ...

        def write(self, value: bytes) -> int:
            return self._text_buffer.write(value.decode())

    def __init__(self, log_func: Callable[[str], None], *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.log_func = log_func
        self.buffer = self._BytesBuffer(self)  # type: ignore[misc,assignment]

    def write(self, message: str) -> int:
        for line in message.splitlines(keepends=False):
            if stripped := line.strip():
                self.log_func(stripped)
        return 0


@contextmanager
def redirect_output_to_logging(stdout_level: str = "info", stderr_level: str = "error") -> Iterator[None]:
    """Redirect standard output and error to logging.

    Yields:
        Nothing.
    """
    with (
        closing(_TextBuffer(getattr(logger, stdout_level))) as new_stdout,
        closing(_TextBuffer(getattr(logger, stderr_level))) as new_stderr,
        redirect_stdout(new_stdout),
        redirect_stderr(new_stderr),
    ):
        yield


def log_captured(text: str, level: str = "info", pkg: str | None = None) -> None:
    """Log captured text.

    Parameters:
        text: The text to split and log.
        level: The log level to use.
    """
    log = getattr(logger, level)
    for line in text.splitlines(keepends=False):
        log(double_brackets(line), pkg=pkg)


def tail(log_file: str) -> None:
    """Tail a log file.

    Parameters:
        log_file: The log file to tail.
    """
    with open(log_file) as file:
        try:
            while True:
                line = file.readline()
                if line:
                    print(line, end="")  # noqa: T201
                else:
                    time.sleep(0.5)
        except KeyboardInterrupt:
            return


def _update_record(record: Record) -> None:
    record["pkg"] = record["extra"].get("pkg") or (record["name"] or "").split(".", 1)[0]  # type: ignore[typeddict-unknown-key]


class _InterceptHandler(logging.Handler):
    def __init__(self, level: int = 0, allow: tuple[str, ...] = ()) -> None:
        super().__init__(level)
        self.allow = allow

    def emit(self, record: logging.LogRecord) -> None:
        # Get corresponding Loguru level if it exists.
        try:
            level: str | int = logger.level(record.levelname).name
        except ValueError:
            level = record.levelno

        # Prevent too much noise from dependencies
        if level == "INFO" and not record.name.startswith(self.allow):
            level = "DEBUG"

        # Find caller from where originated the logged message.
        frame, depth = sys._getframe(6), 6
        while frame and frame.f_code.co_filename == logging.__file__:
            frame = frame.f_back  # type: ignore[assignment]
            depth += 1

        # Log the message, replacing new lines with spaces.
        message = record.getMessage().replace("\n", " ")
        logger.opt(depth=depth, exception=record.exc_info).log(level, message)


intercept_handler = _InterceptHandler()


def configure_logging(
    level: Annotated[str, Doc("Log level (name).")],
    path: Annotated[str | Path | None, Doc("Log file path.")] = None,
    allow: Annotated[
        tuple[str, ...],
        Doc(
            """
            List of package names for which to allow log levels greater or equal to INFO level.
            Packages that are not allowed will see all their logs demoted to DEBUG level.
            If unspecified, allow everything.
            """,
        ),
    ] = (),
) -> None:
    """Configure logging."""
    sink = path or sys.stderr
    log_level = {
        "TRACE": logging.DEBUG - 5,  # 5
        "DEBUG": logging.DEBUG,  # 10
        "INFO": logging.INFO,  # 20
        "SUCCESS": logging.INFO + 5,  # 25
        "WARNING": logging.WARNING,  # 30
        "ERROR": logging.ERROR,  # 40
        "CRITICAL": logging.CRITICAL,  # 50
    }.get(level.upper(), logging.INFO)
    intercept_handler.allow = allow
    logging.basicConfig(handlers=[intercept_handler], level=0, force=True)
    loguru_format = (
        "<green>{time:YYYY-MM-DD HH:mm:ss.SSS}</green> | "
        "<level>{level: <8}</level> | <cyan>{pkg}</cyan> - <level>{message}</level>"
    )
    handler = {"sink": sink, "level": log_level, "format": loguru_format}
    logger.configure(handlers=[handler])  # type: ignore[list-item]


logger = logger.patch(_update_record)



"""Generate the CLI reference in Markdown."""

import getpass
import re

import cappa
import mkdocs_gen_files
from cappa.base import collect
from cappa.help import generate_arg_groups

from insiders._internal.cli import CommandMain

nav = mkdocs_gen_files.Nav()


def _repl_config(match: re.Match) -> str:
    config_key = match.group(1).strip("`")
    config_attr = config_key.replace("-", "_").replace(".", "_")
    return f"[`{config_key}`][insiders._internal.config.Config.{config_attr}]"


def render_parser(command: cappa.Command, title: str, page: str, heading_level: int = 1, layer: int = 1) -> str:
    """Render the parser help documents as a string."""
    result = [f"{'#' * heading_level} **`{title}`**\n"]
    if command.help:
        result.append(f"> {command.help}\n")
    if command.description:
        result.append(f"{command.description}\n")

    for (name, _), args in sorted(generate_arg_groups(command)):
        if name.lower() == "global options" and layer > 1:
            continue
        if name.lower() != "subcommands":
            result.append(f"{name.title()} | Description | Default")
            result.append("--- | --- | ---")
        for arg in args:
            if isinstance(arg, cappa.Subcommand):
                for option in arg.options.values():
                    title = option.real_name()
                    markdown = render_parser(option, title, f"command-{title}.md", heading_level + 1, layer + 1)
                    if layer >= 2:  # noqa: PLR2004
                        result.append(markdown)
                continue

            line = ""
            if name.lower() != "arguments":
                opts = [f"`{opt}`" for opt in arg.names()]
                line += f"`{arg.field_name}`" if not opts else ", ".join(opts)

            line += f" `{arg.value_name.upper()}`" if isinstance(arg.value_name, str) and arg.num_args else ""
            line += f" | {arg.help} | "
            default = arg.show_default.format_default(  # type: ignore[union-attr]
                arg.default,  # type: ignore[arg-type]
                default_format="{default}",
            )
            if default:
                default = re.sub(r"(`(project|github|pypi|backlog|index|polar|team)\.[^`]+`)", _repl_config, default)
                line += default
            result.append(line)
        result.append("")

    markdown = "\n".join(result)

    if layer < 3:  # noqa: PLR2004
        nav[page[:-3]] = page
        markdown = re.sub(rf"\b{re.escape(getpass.getuser())}\b", "user", markdown)
        with mkdocs_gen_files.open(f"cli/{page}", "w") as fd:
            fd.write(markdown)

        mkdocs_gen_files.set_edit_path(f"cli/{page}", "scripts/gen_cli_reference.py")

    return markdown


command = collect(CommandMain, help=False, completion=False)
render_parser(command, "insiders", "index.md")

with mkdocs_gen_files.open("cli/SUMMARY.md", "w") as nav_file:
    nav_file.writelines(nav.build_literate_nav())
    print("\n".join(nav.build_literate_nav()))



from __future__ import annotations

import json
from pathlib import Path
from typing import Annotated as An

from typing_extensions import Doc

from insiders._internal.models import Org, Sponsorship


def update_numbers_file(
    sponsorships: An[list[Sponsorship], Doc("List of sponsorships.")],
    filepath: An[Path, Doc("File-path to update.")] = Path("numbers.json"),
) -> None:
    """Update the file storing sponsorship numbers."""
    with filepath.open("w") as f:
        json.dump(
            {
                "total": sum(sponsorship.amount for sponsorship in sponsorships),
                "count": len(sponsorships),
            },
            f,
            indent=2,
        )


def update_sponsors_file(
    sponsorships: An[list[Sponsorship], Doc("List of sponsorships.")],
    filepath: An[Path, Doc("File-path to update.")] = Path("sponsors.json"),
    *,
    exclude_private: An[bool, Doc("Whether to exclude private members.")] = True,
) -> None:
    """Update the file storing sponsors info."""
    with filepath.open("w") as f:
        json.dump(
            [
                {
                    "name": sponsorship.account.name,
                    "image": sponsorship.account.image,
                    "url": sponsorship.account.url,
                    "org": isinstance(sponsorship.account, Org),
                }
                for sponsorship in sponsorships
                if not sponsorship.private or not exclude_private
            ],
            f,
            indent=2,
        )



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import Annotated as An

from rich.console import Console
from rich.table import Table
from typing_extensions import Doc

from insiders._internal.logger import logger
from insiders._internal.models import Backlog

if TYPE_CHECKING:
    from insiders._internal.clients.github import GitHub
    from insiders._internal.clients.polar import Polar
    from insiders._internal.models import Sponsors


def print_backlog(
    backlog: An[Backlog, Doc("The backlog to print.")],
    labels: An[dict[str, str] | None, Doc("A map of label representations.")] = None,
    limit: An[int, Doc("The maximum number of issues to print.")] = 0,
) -> None:
    """Print the backlog."""
    table = Table(title=f"Backlog ({f'showing {limit} of ' if limit else ''}{len(backlog.issues)} issues)")
    table.add_column("Nº", no_wrap=True)
    table.add_column("Issue", no_wrap=True)
    table.add_column("Author", no_wrap=True)
    table.add_column("Labels", no_wrap=False)
    table.add_column("Funding", no_wrap=True)
    table.add_column("Boost", no_wrap=True)
    table.add_column("Upvotes", no_wrap=True)
    table.add_column("Title")

    labels = labels or {}

    for index, issue in enumerate(backlog.issues, 1):
        iid = f"{issue.repository}#{issue.number}"
        url = f"https://github.com/{issue.repository}/issues/{issue.number}"
        table.add_row(
            str(index),
            f"[link={url}]{iid}[/link]",
            f"[link=https://github.com/{issue.author.name}]{issue.author.name}[/link]",
            "".join(labels.get(label, label) for label in sorted(issue.labels)),
            f"💖{issue.funding}",
            f"💲{issue.pledged}",
            f"👍{len(issue.upvotes)}",
            f"[link={url}]{issue.title}[/link]",
        )
        if index == limit:
            break

    console = Console()
    console.print(table)


def get_backlog(
    github_namespaces: list[str],
    github: GitHub,
    polar: Polar | None = None,
    sponsors: Sponsors | None = None,
    issue_labels: set[str] | None = None,
) -> Backlog:
    github_users = {user for user in sponsors.users if user.platform == "github"} if sponsors else None
    github_issues = github.get_issues(github_namespaces, github_users, allow_labels=issue_labels)
    logger.debug(f"Got {len(github_issues)} issues from GitHub")

    if polar:
        polar_issues = polar.get_issues(github_namespaces, github_users)
        logger.debug(f"Got {len(polar_issues)} issues from Polar")
        for key, github_issue in github_issues.items():
            if key in polar_issues and (polar_issues[key].upvotes or polar_issues[key].pledged):
                github_issue.pledged = polar_issues[key].pledged

    return Backlog(issues=list(github_issues.values()))



"""Default values throughout the project."""

from __future__ import annotations

from pathlib import Path

from platformdirs import user_cache_dir, user_config_dir, user_data_dir

_APP_NAME = "insiders"
_APP_AUTHOR = _APP_NAME

DEFAULT_PORT = 31411
DEFAULT_INDEX_URL = f"http://localhost:{DEFAULT_PORT}"
DEFAULT_REPO_DIR = Path(user_cache_dir(_APP_NAME, _APP_AUTHOR))
DEFAULT_DIST_DIR = Path(user_data_dir(_APP_NAME, _APP_AUTHOR))
DEFAULT_CONF_DIR = Path(user_config_dir(_APP_NAME))
DEFAULT_CONF_PATH = DEFAULT_CONF_DIR / "insiders.toml"



"""Tests for the `config` module."""

from dataclasses import fields

from insiders._internal.config import Config


def test_valid_field_names_and_keys() -> None:
    """Test that all field names match their unset name."""
    for field in fields(Config):
        assert field.name == field.default.name



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import Annotated as An

import httpx
from typing_extensions import Doc

from insiders._internal.clients import Client

if TYPE_CHECKING:
    from collections.abc import Iterable, Mapping

    from insiders._internal.models import Sponsors


class KoFi(Client):
    """Ko-fi client."""

    name = "Ko-fi"

    # NOTE: Ko-fi doesn't seem to provide an HTTP API, only webhooks...
    def __init__(self, token: An[str, Doc("A KoFi token.")]) -> None:  # noqa: ARG002
        """Initialize KoFi API client."""
        self.http_client = httpx.Client()

    def get_sponsors(
        self,
        org_members_map: Mapping[str, Iterable[str]] | None = None,
        *,
        exclude_private: bool = False,
    ) -> An[Sponsors, Doc("Sponsors data.")]:
        """Get Ko-fi sponsors."""
        raise NotImplementedError("Ko-fi support is not implemented yet")



"""insiders package.

Manage your Insiders projects.
"""

from __future__ import annotations

from insiders._internal.cli import main
from insiders._internal.clients.github import GitHub
from insiders._internal.clients.index import Index
from insiders._internal.clients.polar import Polar
from insiders._internal.models import Account, Backlog, Issue, IssueDict, Org, Sponsors, Sponsorship, User
from insiders._internal.ops.backlog import get_backlog, print_backlog

__all__: list[str] = [
    "Account",
    "Backlog",
    "GitHub",
    "Index",
    "Issue",
    "IssueDict",
    "Org",
    "Polar",
    "Sponsors",
    "Sponsorship",
    "User",
    "get_backlog",
    "main",
    "print_backlog",
]



from __future__ import annotations

from functools import partial
from pathlib import Path
from typing import Annotated as An
from typing import Any

from copier import run_copy as copier_run
from pypi_insiders.logger import run as run_and_log
from typing_extensions import Doc

from insiders._internal.logger import logger

# TODO: Handle more operations:
# - Create matching Insiders project from public one.
# - Create matching public project from Insiders one.
# - What else? Get user feedback.

_gh_repo_create = partial(run_and_log, "gh", "repo", "create")


def new_public_and_insiders_github_projects(
    *,
    public_namespace: An[str, Doc("Namespace of the public repository.")],
    public_name: An[str, Doc("Name of the public repository.")],
    description: An[str, Doc("Shared description.")],
    public_repo_path: An[str | Path, Doc("Local path in which to clone the public repository.")],
    insiders_namespace: An[
        str | None,
        Doc("Namespace of the insiders repository. Defaults to the public namespace."),
    ] = None,
    insiders_name: An[str | None, Doc("Name of the insiders repository. Defaults to the public name.")] = None,
    insiders_repo_path: An[str | Path, Doc("Local path in which to clone the insiders repository.")],
    github_username: An[str | None, Doc("Username. Defaults to the public namespace value.")] = None,
    copier_template: An[str | None, Doc("Copier template to initialize the local insiders repository with.")] = None,
    copier_template_answers: An[dict[str, Any] | None, Doc("Answers to the Copier template questions.")] = None,
    post_creation_command: An[
        str | list[str] | None,
        Doc("Command to run after creating the public repository."),
    ] = None,
) -> None:
    """Create a new Insiders project on GitHub (public and private repositories)."""
    github_username = github_username or public_namespace
    insiders_namespace = insiders_namespace or f"{github_username}-insiders"
    insiders_name = insiders_name or public_name
    github_description = f"{description} Available to sponsors only."

    logger.debug("Creating new project with these settings:")
    logger.debug(f"- public repo:   {public_namespace}/{public_name} cloned in {public_repo_path}")
    logger.debug(f"- insiders repo: {insiders_namespace}/{insiders_name} cloned in {insiders_repo_path}")

    common_opts = ("--disable-wiki", "--homepage", f"https://{public_namespace}.github.io/{public_name}")
    public_opts = ("--description", github_description, "--public", *common_opts)
    insiders_opts = ("--description", description, "--private", "--disable-issues", *common_opts)
    _gh_repo_create(f"{public_namespace}/{public_name}", *public_opts)
    _gh_repo_create(f"{insiders_namespace}/{insiders_name}", *insiders_opts)

    repo_path = Path(public_repo_path)
    run_and_log("git", "clone", f"git@github.com:{public_namespace}/{public_name}", repo_path)

    if copier_template:
        context = {
            "public_name": public_name,
            "public_namespace": public_namespace,
            "insiders_name": insiders_name,
            "insiders_namespace": insiders_namespace,
            "description": description,
        }
        formatted_answers = {
            key: value.format(**context) if isinstance(value, str) else value
            for key, value in (copier_template_answers or {}).items()
        }
        copier_run(
            copier_template,
            repo_path,
            user_defaults=formatted_answers,
            overwrite=True,
            unsafe=True,
        )

    commit_message = f"feat: Generate project with {copier_template} Copier template"
    run_and_log("git", "-C", repo_path, "add", "-A")
    run_and_log("git", "-C", repo_path, "commit", "-m", commit_message)

    if isinstance(post_creation_command, str):
        run_and_log(post_creation_command, cwd=repo_path, shell=True)  # noqa: S604
    elif isinstance(post_creation_command, list):
        run_and_log(*post_creation_command, cwd=repo_path)
    else:
        run_and_log("git", "-C", repo_path, "push")

    insiders_repo_path = Path(insiders_repo_path)
    run_and_log("git", "clone", f"git@github.com:{insiders_namespace}/{insiders_name}", insiders_repo_path)
    run_and_log(
        "git",
        "-C",
        insiders_repo_path,
        "remote",
        "add",
        "upstream",
        f"git@github.com:{public_namespace}/{public_name}",
    )
    run_and_log("git", "-C", insiders_repo_path, "pull", "upstream", "main")



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import Annotated as An

from typing_extensions import Doc, Self

if TYPE_CHECKING:
    from collections.abc import Iterable, Mapping
    from types import TracebackType

    import httpx

    from insiders._internal.models import Sponsors


class Client:
    name: An[str, Doc("Name of the platform.")]
    http_client: An[httpx.Client, Doc("""An HTTPX client.""")]

    def __enter__(self) -> Self:
        """Enter context manager."""
        self.http_client.__enter__()
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None = None,
        exc_value: BaseException | None = None,
        traceback: TracebackType | None = None,
    ) -> None:
        """Exit context manager."""
        self.http_client.__exit__(exc_type, exc_value, traceback)

    def get_sponsors(
        self,
        org_members_map: An[
            Mapping[str, Iterable[str]] | None,
            Doc("A mapping of organization name to members."),
        ] = None,
    ) -> Sponsors:
        raise NotImplementedError("Not implemented")



"""Tests for the `cli` module."""

from __future__ import annotations

from typing import TYPE_CHECKING

from insiders._internal import cli, debug

if TYPE_CHECKING:
    import pytest


def test_main() -> None:
    """Basic CLI test."""
    assert cli.main([]) == 2


def test_show_help(capsys: pytest.CaptureFixture) -> None:
    """Show help.

    Parameters:
        capsys: Pytest fixture to capture output.
    """
    assert cli.main(["-h"]) == 0
    captured = capsys.readouterr()
    assert "insiders" in captured.out


def test_show_version(capsys: pytest.CaptureFixture) -> None:
    """Show version.

    Parameters:
        capsys: Pytest fixture to capture output.
    """
    assert cli.main(["-V"]) == 0
    captured = capsys.readouterr()
    assert debug.get_version() in captured.out


def test_show_debug_info(capsys: pytest.CaptureFixture) -> None:
    """Show debug information.

    Parameters:
        capsys: Pytest fixture to capture output.
    """
    assert cli.main(["--debug-info"]) == 0
    captured = capsys.readouterr().out.lower()
    assert "python" in captured
    assert "system" in captured
    assert "environment" in captured
    assert "packages" in captured



from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING
from typing import Annotated as An

import httpx
from typing_extensions import Doc

from insiders._internal.clients import Client
from insiders._internal.logger import logger
from insiders._internal.models import Issue, IssueDict, Org, Sponsors, Sponsorship, User

if TYPE_CHECKING:
    from collections.abc import Iterable, Mapping


_GRAPHQL_SPONSORS_QUERY = """
query {
    viewer {
        sponsorshipsAsMaintainer(
            first: 100
            after: %s
            includePrivate: true
            orderBy: {
                field: CREATED_AT
                direction: DESC
            }
        )
        {
            pageInfo {
                hasNextPage
                endCursor
            }
            nodes {
                createdAt
                isOneTimePayment
                privacyLevel
                sponsorEntity {
                    ...on Actor {
                        __typename
                        login
                        avatarUrl
                        url
                    }
                },
                tier {
                    monthlyPriceInDollars
                }
            }
        }
    }
}
"""

_GRAPHQL_ISSUES_QUERY = """
query {
    search(
        first: 100
        after: %(after)s
        query: "%(query)s"
        type: ISSUE
    )
    {
        pageInfo {
            hasNextPage
            endCursor
        }
        nodes {
            __typename
            ... on Issue {
                author {
                    login
                }
                title
                number
                repository {
                    nameWithOwner
                }
                createdAt
                labels(first: 10) {
                    nodes {
                        name
                    }
                }
                reactions(first: 100) {
                    nodes {
                        content
                        user {
                            login
                        }
                    }
                }
            }
        }
    }
}
"""


class GitHub(Client):
    """GitHub client."""

    name = "GitHub"

    def __init__(
        self,
        token: An[str, Doc("""A GitHub token. Recommended scopes: `admin:org` and `read:user`.""")],
    ) -> None:
        """Initialize GitHub API client."""
        self.http_client = httpx.Client(
            base_url="https://api.github.com",
            headers={"Authorization": f"Bearer {token}"},
        )

    def get_org_members(
        self,
        org: An[str, Doc("The organization name.")],
    ) -> An[set[str], Doc("A set of member names.")]:
        """Get organization members (username only)."""
        page = 1
        members = set()
        while True:
            response = self.http_client.get(f"/orgs/{org}/members", params={"per_page": 100, "page": page})
            response.raise_for_status()
            response_data = response.json()
            for member in response_data:
                members.add(member["login"])
            if len(response_data) < 100:  # noqa: PLR2004
                break
            page += 1
        return members

    def get_sponsors(
        self,
        org_members_map: An[
            Mapping[str, Iterable[str]] | None,
            Doc("A mapping of organization name to members."),
        ] = None,
        *,
        exclude_private: bool = False,
    ) -> An[Sponsors, Doc("Sponsors data.")]:
        """Get GitHub sponsors."""
        logger.debug("Fetching sponsors from GitHub.")
        sponsorships = []
        user_accounts = {}
        cursor = "null"

        while True:
            # Get sponsors data.
            logger.debug(f"Fetching page of sponsors from GitHub with cursor {cursor}.")
            payload = {"query": _GRAPHQL_SPONSORS_QUERY % cursor}
            response = self.http_client.post("/graphql", json=payload)
            response.raise_for_status()

            # Process sponsors data.
            data = response.json()["data"]
            for item in data["viewer"]["sponsorshipsAsMaintainer"]["nodes"]:
                if item["isOneTimePayment"]:
                    continue
                private = item["privacyLevel"].lower() == "private"
                if private and exclude_private:
                    continue

                # Determine account.
                account: User | Org
                account_data = {
                    "name": item["sponsorEntity"]["login"],
                    "image": item["sponsorEntity"]["avatarUrl"],
                    "url": item["sponsorEntity"]["url"],
                    "platform": "github",
                }
                if item["sponsorEntity"]["__typename"].lower() == "organization":
                    account = Org(**account_data)
                    logger.debug(f"Found org: @{account.name}")
                else:
                    account = User(**account_data)
                    logger.debug(f"Found user: @{account.name}")
                    user_accounts[account.name] = account

                # Record sponsorship.
                sponsorships.append(
                    Sponsorship(
                        private=private,
                        created=datetime.strptime(item["createdAt"], "%Y-%m-%dT%H:%M:%SZ"),  # noqa: DTZ007
                        amount=item["tier"]["monthlyPriceInDollars"],
                        account=account,
                    ),
                )

            # Check for next page.
            if data["viewer"]["sponsorshipsAsMaintainer"]["pageInfo"]["hasNextPage"]:
                cursor = f'"{data["viewer"]["sponsorshipsAsMaintainer"]["pageInfo"]["endCursor"]}"'
            else:
                break

        # Consolidate data.
        logger.debug(f"Processing {len(sponsorships)} sponsorships from GitHub.")
        for sponsorship in sponsorships:
            # Accounts link back to their sponsorship.
            sponsorship.account.sponsorships.append(sponsorship)
            # Add users as org members, link their org sponsorship.
            if sponsorship.account.is_org and org_members_map and sponsorship.account.name in org_members_map:
                org_members = org_members_map[sponsorship.account.name]
                actual_org_members = self.get_org_members(sponsorship.account.name)
                for org_member in org_members:
                    if org_member not in user_accounts:
                        user_accounts[org_member] = User(name=org_member, platform="github")
                    org_member_account = user_accounts[org_member]
                    verified = org_member in actual_org_members
                    sponsorship.account.users[org_member] = (org_member_account, verified)  # type: ignore[union-attr]
                    logger.debug(f"Found org member: @{sponsorship.account.name}/{org_member} (verified: {verified})")
                    org_member_account.sponsorships.append(sponsorship)

        return Sponsors(sponsorships=sponsorships)

    def get_team_members(
        self,
        org: An[str, Doc("The organization name.")],
        team: An[str, Doc("The team name.")],
    ) -> An[set[str], Doc("A set of member names.")]:
        """Get members of a GitHub team."""
        logger.debug(f"Fetching members of {org}/{team} team.")
        page = 1
        members = set()
        while True:
            response = self.http_client.get(f"/orgs/{org}/teams/{team}/members", params={"per_page": 100, "page": page})
            response.raise_for_status()
            response_data = response.json()
            members |= {member["login"] for member in response_data}
            if len(response_data) < 100:  # noqa: PLR2004
                break
            page += 1
        return members

    def get_team_invites(
        self,
        org: An[str, Doc("The organization name.")],
        team: An[str, Doc("The team name.")],
    ) -> An[set[str], Doc("A set of member names.")]:
        """Get pending invitations to a GitHub team."""
        logger.debug(f"Fetching pending invitations to {org}/{team} team.")
        page = 1
        invites = set()
        while True:
            response = self.http_client.get(f"/orgs/{org}/teams/{team}/invitations", params={"per_page": 100})
            response.raise_for_status()
            response_data = response.json()
            invites |= {invite["login"] for invite in response_data}
            if len(response_data) < 100:  # noqa: PLR2004
                break
            page += 1
        return invites

    def get_failed_invites(
        self,
        org: An[str, Doc("The organization name.")],
    ) -> An[set[str], Doc("A set of member names.")]:
        logger.debug(f"Fetching failed invitations to {org} organization.")
        page = 1
        invites = set()
        while True:
            response = self.http_client.get(f"/orgs/{org}/failed_invitations", params={"per_page": 100})
            response.raise_for_status()
            response_data = response.json()
            invites |= {invite["login"] for invite in response_data}
            if len(response_data) < 100:  # noqa: PLR2004
                break
            page += 1
        return invites

    def grant_access(
        self,
        user: An[str, Doc("A username.")],
        org: An[str, Doc("An organization name.")],
        team: An[str, Doc("A team name.")],
    ) -> None:
        """Grant access to a user to a GitHub team."""
        logger.debug(f"Granting @{user} access to {org}/{team} team.")
        response = self.http_client.put(f"/orgs/{org}/teams/{team}/memberships/{user}")
        response.raise_for_status()
        # try:
        #     response.raise_for_status()
        # except httpx.HTTPError as error:
        #     logger.error(f"Couldn't add @{user} to {org}/{team} team: {error}")
        #     if response.content:
        #         response_body = response.json()
        #         logger.error(f"{response_body['message']} See {response_body['documentation_url']}")
        # else:
        #     logger.info(f"@{user} added to {org}/{team} team")

    def revoke_access(
        self,
        user: An[str, Doc("A username.")],
        org: An[str, Doc("An organization name.")],
        team: An[str, Doc("A team name.")],
    ) -> None:
        """Revoke access from a user to a GitHub team."""
        logger.debug(f"Revoking access from @{user} to {org}/{team} team.")
        response = self.http_client.delete(f"/orgs/{org}/teams/{team}/memberships/{user}")
        response.raise_for_status()
        # try:
        #     response.raise_for_status()
        # except httpx.HTTPError as error:
        #     logger.error(f"Couldn't remove @{user} from {org}/{team} team: {error}")
        #     if response.content:
        #         response_body = response.json()
        #         logger.error(f"{response_body['message']} See {response_body['documentation_url']}")
        # else:
        #     logger.info(f"@{user} removed from {org}/{team} team")

    def get_issues(
        self,
        github_accounts: An[Iterable[str], Doc("A list of GitHub account names.")],
        known_github_users: An[Iterable[User] | None, Doc("Known user accounts.")] = None,
        *,
        allow_labels: An[set[str] | None, Doc("A set of labels to keep.")] = None,
    ) -> An[IssueDict, Doc("A dictionary of issues.")]:
        """Get issues from GitHub."""
        logger.debug("Fetching issues from GitHub.")

        known_users = {account.name: account for account in (known_github_users or ())}
        issues = {}
        allow_labels = allow_labels or set()
        cursor = "null"
        users_query = " ".join(f"user:{user}" for user in github_accounts)
        query = f"{users_query} sort:created state:open"

        while True:
            # Get issues data.
            logger.debug(f"Fetching page of issues from GitHub with cursor {cursor}.")
            payload = {"query": _GRAPHQL_ISSUES_QUERY % {"after": cursor, "query": query}}
            response = self.http_client.post("/graphql", json=payload)
            response.raise_for_status()

            # Process issues data.
            data = response.json()["data"]
            for issue in data["search"]["nodes"]:
                if issue["__typename"] != "Issue":
                    continue
                author_id = issue["author"]["login"].removesuffix("[bot]")
                repository = issue["repository"]["nameWithOwner"]
                title = issue["title"]
                number = issue["number"]
                created_at = datetime.strptime(issue["createdAt"], "%Y-%m-%dT%H:%M:%SZ")  # noqa: DTZ007
                labels = {label["name"] for label in issue["labels"]["nodes"] if label["name"] in allow_labels}

                if author_id not in known_users:
                    known_users[author_id] = User(name=author_id, platform="github")
                author = known_users[author_id]

                upvotes = set()
                for reaction in issue["reactions"]["nodes"]:
                    if reaction["content"] == "THUMBS_UP":
                        upvoter_id = reaction["user"]["login"]
                        if upvoter_id not in known_users:
                            known_users[upvoter_id] = User(name=upvoter_id, platform="github")
                        upvoter = known_users[upvoter_id]
                        upvotes.add(upvoter)

                iid = (repository, number)
                issues[iid] = Issue(
                    repository=repository,
                    number=number,
                    title=title,
                    created=created_at,
                    author=author,
                    upvotes=upvotes,
                    labels=labels,
                    platform="github",
                )

            # Check for next page.
            if data["search"]["pageInfo"]["hasNextPage"]:
                cursor = f'"{data["search"]["pageInfo"]["endCursor"]}"'
            else:
                break

        return issues

    def sync_team(
        self,
        team: An[str, Doc("GitHub team to sync sponsors with.")],
        *,
        sponsors: An[Sponsors | None, Doc("Sponsors data.")] = None,
        min_amount: An[int | None, Doc("Minimum amount to be considered a sponsor.")] = None,
        include_users: An[set[str] | None, Doc("Users to always grant access to.")] = None,
        exclude_users: An[set[str] | None, Doc("Users to never grant access to.")] = None,
        org_users: An[dict[str, set[str]] | None, Doc("Users to grant access to based on org.")] = None,
    ) -> None:
        """Sync sponsors with members of a GitHub team."""
        sponsors = sponsors or self.get_sponsors(org_users)

        eligible_users = {user.name for user in sponsors.users if not min_amount or user.tier_sum >= min_amount}
        if include_users:
            eligible_users |= include_users
        if exclude_users:
            eligible_users -= exclude_users

        org, team = team.split("/", 1)
        invitable_users = eligible_users - self.get_failed_invites(org)
        members = self.get_team_members(org, team) | self.get_team_invites(org, team)

        # Revoke accesses.
        for user in members:
            if user not in eligible_users:
                self.revoke_access(user, org, team)

        # Grant accesses.
        for user in invitable_users:
            if user not in members:
                self.grant_access(user, org, team)

    def create_repo(
        self,
        repository: An[str, Doc("The repository, like `namespace/repo`.")],
        *,
        description: An[str | None, Doc("The repository description.")] = None,
        homepage: An[str | None, Doc("The repository homepage.")] = None,
        private: An[bool, Doc("Whether the repository is private.")] = False,
        has_issues: An[bool, Doc("Enable issues.")] = False,
        has_projects: An[bool, Doc("Enable projects.")] = False,
        has_wiki: An[bool, Doc("Enable the wiki.")] = False,
        has_discussions: An[bool, Doc("Enable discussions.")] = False,
    ) -> None:
        """Create a repository."""
        # NOTE: No way to create discussion categories via API.
        logger.debug(f"Creating repository {repository}.")

        # Determine account type.
        try:
            account, repo_name = repository.split("/")
        except ValueError:
            repo_name = repository
            url = "/user/repos"
        else:
            response = self.http_client.get(f"/users/{account}")
            response.raise_for_status()
            response_data = response.json()
            url = f"/orgs/{account}/repos" if response_data["type"] == "Organization" else "/user/repos"

        # Create the repository.
        response = self.http_client.post(
            url,
            params={
                "name": repo_name,
                "description": description,
                "homepage": homepage,
                "private": private,
                "has_issues": has_issues,
                "has_projects": has_projects,
                "has_wiki": has_wiki,
                "has_discussions": has_discussions,
            },
        )
        response.raise_for_status()



"""Manage PyPI-like index."""

from __future__ import annotations

import os
import shutil
import sys
from collections.abc import Iterator
from pathlib import Path
from typing import TYPE_CHECKING
from typing import Annotated as An
from urllib.parse import urlparse

import psutil
from build.__main__ import build_package
from failprint import Capture
from packaging.version import InvalidVersion, Version
from pypiserver.__main__ import main as serve
from twine.commands.upload import upload
from twine.settings import Settings
from typing_extensions import Doc
from unearth import PackageFinder

from insiders._internal import defaults
from insiders._internal.logger import log_captured, logger, redirect_output_to_logging, run

# YORE: EOL 3.10: Replace block with line 2.
if sys.version_info >= (3, 11):
    import tomllib
else:
    import tomli as tomllib

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator
    from typing import Any


def _normalize_version(version: str) -> str:
    if version[0] == "v":
        version = version[1:]
    return version.replace("+", ".").replace("-", ".")


class GitCache:
    """A cache for local clones of configured repositories."""

    def __init__(self, cache_dir: str | Path = defaults.DEFAULT_REPO_DIR) -> None:
        """Initialize the cache.

        Parameters:
            cache_dir: The directory in which to clone the repositories.
        """
        self.cache_dir: Path = Path(cache_dir)
        self.cache_dir.mkdir(exist_ok=True, parents=True)

    def _git(self, repo: str, *args: str, **kwargs: Any) -> str:
        cached_repo = self.cache_dir / repo
        return run("git", "-C", cached_repo, *args, **kwargs)

    def list(self) -> Iterator[str]:
        """List the repositories in the cache.

        Returns:
            An iterator over the repository names.
        """
        yield from (repo.name for repo in self.cache_dir.iterdir())

    def exists(self, repo: str) -> bool:
        """Check if a repository already exists.

        Parameters:
            repo: The repository to check.

        Returns:
            True or false.
        """
        return self.cache_dir.joinpath(repo).exists()

    def clone(self, repo: str, url: str) -> Path:
        """Clone a repository.

        Parameters:
            repo: The repository to clone.

        Returns:
            The path to the cloned repository.
        """
        logger.debug(f"{repo}: Cloning {url}")
        cached_repo = self.cache_dir / repo
        cached_repo.parent.mkdir(exist_ok=True)
        run("git", "clone", url, cached_repo)
        return cached_repo

    def checkout(self, repo: str, ref: str) -> None:
        """Checkout a ref.

        Parameters:
            repo: The repository to work on.
        """
        logger.debug(f"{repo}: Checking out {ref}")
        self._git(repo, "checkout", ref)

    def checkout_origin_head(self, repo: str) -> None:
        """Checkout origin's HEAD again.

        Parameters:
            repo: The repository to work on.
        """
        logger.debug(f"{repo}: Checking out origin's HEAD")
        self._git(repo, "remote", "set-head", "origin", "--auto")
        ref = self._git(repo, "symbolic-ref", "refs/remotes/origin/HEAD")
        self.checkout(repo, ref.strip().split("/")[3])

    def pull(self, repo: str) -> None:
        """Pull latest changes.

        Parameters:
            repo: The repository to work on.
        """
        logger.debug(f"{repo}: Pulling latest changes")
        self._git(repo, "pull")

    def dist_name(self, repo: str) -> str:
        """Get the distribution name.

        Parameters:
            repo: The repository to work on.

        Returns:
            The distribution name.
        """
        with self.cache_dir.joinpath(repo, "pyproject.toml").open("rb") as file:
            return tomllib.load(file)["project"]["name"]

    def latest_tag(self, repo: str) -> str:
        """Get the latest Git tag.

        Parameters:
            repo: The repository to work on.

        Returns:
            A tag.
        """
        return self._git(repo, "describe", "--tags", "--abbrev=0").strip()

    def remove(self, repo: str) -> None:
        """Remove a repository from the cache.

        Parameters:
            repo: The repository to remove.
        """
        shutil.rmtree(self.cache_dir / repo, ignore_errors=True)

    def remove_dist(self, repo: str) -> None:
        """Remove the `dist` folder of a repository.

        Parameters:
            repo: The repository to work on.
        """
        shutil.rmtree(self.cache_dir.joinpath(repo, "dist"), ignore_errors=True)

    def build(self, repo: str) -> Iterator[Path]:
        """Build distributions.

        Parameters:
            repo: The repository to work on.

        Returns:
            File path for each distribution.
        """
        cached_repo = self.cache_dir / repo
        with Capture.BOTH.here() as captured:
            build_package(cached_repo, cached_repo / "dist", distributions=["sdist", "wheel"])
        log_captured(str(captured), level="debug", pkg="build")
        return cached_repo.joinpath("dist").iterdir()


class Index:
    """Index of repositories."""

    def __init__(
        self,
        url: str = defaults.DEFAULT_INDEX_URL,
        git_dir: Path = defaults.DEFAULT_REPO_DIR,
        dist_dir: Path = defaults.DEFAULT_DIST_DIR,
    ) -> None:
        """Initialize the index.

        Parameters:
            conf_path: The path to the configuration file.
        """
        self.url: str = url
        self.git_dir: Path = git_dir
        self.dist_dir: Path = dist_dir

        parsed = urlparse(url)
        self.port = parsed.port or 80

        self.dist_dir.mkdir(parents=True, exist_ok=True)
        self._git_cache = GitCache(self.git_dir)
        self._finder = PackageFinder(index_urls=[f"{self.url}/simple"])

    def add(self, git_url: str, repo: str | None = None) -> None:
        repo = repo or git_url.split("/")[-1].replace(".git", "")
        cache = self._git_cache
        if not cache.exists(repo):
            cache.clone(repo, git_url)
            cache.checkout(repo, cache.latest_tag(repo))
            self.upload(cache.build(repo))

    def remove(self, repo: str) -> None:
        try:
            dist_name = self._git_cache.dist_name(repo).replace("-", "_")
        except FileNotFoundError:
            logger.warning(f"{repo}: Repository not found, skipping")
            return
        self._git_cache.remove(repo)
        for dist in self.dist_dir.glob(f"{dist_name}-*"):
            dist.unlink()

    def list_distributions(self) -> Iterator[Path]:
        yield from self.dist_dir.iterdir()

    def list_projects(self) -> Iterator[Path]:
        yield from self._git_cache.cache_dir.iterdir()

    def update(self, projects: Iterable[str] | None = None) -> None:
        """Update PyPI packages.

        For each configured repository, pull latest contents,
        checkout latest tag, and if the corresponding version is not present on the index,
        build and upload distributions.
        """
        cache = self._git_cache
        projects = projects or sorted(cache.list())
        for name in projects:
            if not cache.exists(name):
                logger.warning(f"{name}: Repository not found, skipping")
                continue
            dist_name = cache.dist_name(name)
            logger.info(f"{name}: Updating sources")
            cache.checkout_origin_head(name)
            cache.pull(name)

            latest_tag = cache.latest_tag(name)
            if latest_tag:
                logger.debug(f"{name}: Latest tag is {latest_tag}")
            else:
                logger.debug(f"{name}: No tags found")
                continue
            latest_version = self.latest(dist_name)
            if latest_version:
                logger.debug(f"{name}: Latest PyPI version is {latest_version}")

            normal_tag = _normalize_version(latest_tag)
            normal_version = _normalize_version(latest_version or "0.0.0")
            if latest_version is None or normal_tag != normal_version:
                try:
                    if Version(normal_tag) < Version(normal_version):
                        logger.warning(f"Latest tag {latest_tag} is older than latest PyPI version {latest_version}")
                        if self.exists(dist_name, normal_tag):
                            continue
                except InvalidVersion:
                    pass
                logger.info(f"{name}: Building and uploading {normal_tag} (current: {normal_version})")
                cache.remove_dist(name)
                cache.checkout(name, latest_tag)
                new_dists = cache.build(name)
                self.upload(new_dists)
                logger.success(f"{name}: Built and published version {normal_tag}")

    def start(self, *, background: bool = False, log_path: str | None = None) -> None:
        """Start the server."""
        if background and os.fork():
            return
        args = [
            "run",
            str(self.dist_dir),
            f"-p{self.port}",
            "-a.",
            "-P.",
            "--log-stream=none",
        ]
        if log_path:
            args.extend(("--log-file", log_path))
        serve(args)

    def stop(self) -> An[bool, Doc("Whether the server was stopped or not.")]:
        """Stop the server."""
        for proc in psutil.process_iter():
            try:
                cmdline = " ".join(proc.cmdline())
            except psutil.Error:
                continue
            if "insiders index start" in cmdline:
                proc.kill()
                return True
        return False

    def status(self) -> An[dict | None, Doc("Some metadata about the server process.")]:
        """Return the server status as a dict of metadata."""
        for proc in psutil.process_iter():
            try:
                cmdline = " ".join(proc.cmdline())
            except psutil.Error:
                continue
            if "insiders index start" in cmdline:
                return proc.as_dict(attrs=("ppid", "create_time", "username", "name", "cmdline", "pid"))
        return None

    def logs(self) -> str:
        """Return the logs file path."""
        if not (status := self.status()):
            raise FileNotFoundError("Server not running")
        for index, arg in enumerate(status["cmdline"][1:], 1):
            if arg.startswith("-p"):
                if len(arg) > 2:  # noqa: PLR2004
                    return arg[2:]
                return status["cmdline"][index + 1]
            if arg.startswith("--server-log-path"):
                if len(arg) > 18:  # noqa: PLR2004
                    return arg[18:]
                return status["cmdline"][index + 1]
        raise FileNotFoundError("Logs file not found")

    def latest(
        self,
        name: An[str, Doc("The package name (distribution name).")],
    ) -> An[str | None, Doc("The version as a string, or none.")]:
        """Get the latest version of a package."""
        result = self._finder.find_best_match(name, allow_prereleases=True, allow_yanked=True)
        return result.best.version if result.best else None

    def exists(
        self,
        name: An[str, Doc("The package name (distribution name).")],
        version: An[str, Doc("The package version.")],
    ) -> An[bool, Doc("Whether the package version exists or not.")]:
        """Tell if a package version exists."""
        result = self._finder.find_best_match(f"{name}=={version}", allow_prereleases=True, allow_yanked=True)
        return bool(result.best)

    def upload(self, dists: An[Iterable[str | Path], Doc("The distributions to upload.")]) -> None:
        """Upload distributions."""
        with redirect_output_to_logging(stdout_level="debug"):
            upload(
                Settings(
                    non_interactive=True,
                    skip_existing=True,
                    repository_url=self.url,
                    username="",
                    password="",
                    disable_progress_bar=True,
                    verbose=True,
                ),
                [str(dist) for dist in dists],
            )



from __future__ import annotations

import subprocess
import sys
from dataclasses import dataclass, fields
from dataclasses import field as dataclass_field
from typing import TYPE_CHECKING, Any, overload

from insiders._internal.defaults import DEFAULT_CONF_PATH
from insiders._internal.logger import logger
from insiders._internal.models import Backlog  # noqa: F401

if TYPE_CHECKING:
    from collections.abc import Callable
    from pathlib import Path

# YORE: EOL 3.10: Replace block with line 2.
if sys.version_info >= (3, 11):
    import tomllib
else:
    import tomli as tomllib


class Unset:
    """A sentinel value for unset configuration options."""

    def __init__(self, key: str, transform: str | None = None) -> None:
        self.key = key
        self.name = key.replace("-", "_").replace(".", "_")
        self.transform = transform

    def __bool__(self) -> bool:
        return False

    def __repr__(self) -> str:
        return f"<Unset({self.name!r})>"

    def __str__(self) -> str:
        # The string representation is used in the CLI, to show the default values.
        return f"`{self.key}` config-value"


def config_field(key: str, transform: str | None = None) -> Unset:
    """Get a dataclass field with a TOML key."""
    return dataclass_field(default=Unset(key, transform=transform))


@dataclass(kw_only=True)
class Config:
    """Configuration for the insiders project."""

    # Sponsors fields.
    sponsors_minimum_amount: int | Unset = config_field("sponsors.minimum-amount")
    """Minimum sponsorship amount to be considered an insider."""

    # Backlog fields.
    backlog_namespaces: list[str] | Unset = config_field("backlog.namespaces")
    """GitHub namespaces to fetch issues from."""
    backlog_sort: list[Callable] | Unset = config_field("backlog.sort", transform="_eval_sort")
    """Sort strategies to apply to the backlog."""
    backlog_limit: int | Unset = config_field("backlog.limit")
    """Limit the number of displayed issues."""
    backlog_issue_labels: dict[str, str] | Unset = config_field("backlog.issue-labels")
    """Map of label names to their display value (emojis, typically)."""

    # GitHub fields.
    github_token_command: str | Unset = config_field("github.token-command")
    """Command to obtain a GitHub token."""
    github_organization_members: dict[str, set[str]] | Unset = config_field("github.organization-members")
    """Map of organization names to their members."""
    github_project_namespace: str | Unset = config_field("github.project-namespace")
    """GitHub namespace to create public projects in."""
    github_insiders_project_namespace: str | Unset = config_field("github.insiders-project-namespace")
    """GitHub namespace to create insiders projects in."""
    github_username: str | Unset = config_field("github.username")
    """GitHub username to use for operations."""
    github_insiders_team: str | Unset = config_field("github.insiders-team")
    """GitHub team to add insiders to."""
    github_sponsored_account: str | Unset = config_field("github.sponsored-account")
    """GitHub account to use for sponsored operations."""
    github_include_users: set[str] | Unset = config_field("github.include-users")
    """GitHub users to always include in the insiders team (even non-sponsors)."""
    github_exclude_users: set[str] | Unset = config_field("github.exclude-users")
    """GitHub users to never include in the insiders team (even sponsors)."""

    # Project fields.
    project_copier_template: str | Unset = config_field("project.copier-template")
    """Copier template to generate new projects with."""
    project_copier_template_answers: dict[str, str] | Unset = config_field("project.copier-template-answers")
    """Copier template answers to use when generating a project."""
    project_register_on_pypi: bool | Unset = config_field("project.register-on-pypi")
    """Whether to register new projects on PyPI after creating them."""
    project_directory: str | Unset = config_field("project.directory")
    """Directory in which to clone created public projects."""
    project_insiders_directory: str | Unset = config_field("project.insiders-directory")
    """Directory in which to clone created private projects."""
    project_post_creation_command: str | list[str] | Unset = config_field("project.post-creation-command")
    """Command to run after creating a project."""

    # PyPI fields.
    pypi_username: str | Unset = config_field("pypi.username")
    """PyPI username to use for operations."""

    # Index fields.
    index_distributions_directory: str | Unset = config_field("index.distributions-directory")
    """Directory to store generated Python project distributions."""
    index_sources_directory: str | Unset = config_field("index.sources-directory")
    """Directory to store project sources (cloned repositories)."""
    index_url: str | Unset = config_field("index.url")
    """URL of the index server."""
    index_start_in_background: bool | Unset = config_field("index.start-in-background")
    """Whether to start the index server in the background."""
    index_log_path: str | Unset = config_field("index.log-path")
    """Where to write the index server logs to."""

    # Polar fields.
    polar_token_command: str | Unset = config_field("polar.token-command")
    """Command to obtain a Polar token."""
    polar_sponsored_account: str | Unset = config_field("polar.sponsored-account")
    """Polar account to use for sponsored operations."""

    @property
    def github_token(self) -> str | Unset:
        """Get the GitHub token."""
        if isinstance(self.github_token_command, Unset):
            return self.github_token_command
        return subprocess.getoutput(self.github_token_command)  # noqa: S605

    @property
    def polar_token(self) -> str | Unset:
        """Get the Polar token."""
        if isinstance(self.polar_token_command, Unset):
            return self.polar_token_command
        return subprocess.getoutput(self.polar_token_command)  # noqa: S605

    @overload
    @staticmethod
    def _eval_sort(strats: Unset) -> Unset: ...

    @overload
    @staticmethod
    def _eval_sort(strats: list[str]) -> list[Callable]: ...

    @staticmethod
    def _eval_sort(strats: list[str] | Unset) -> list[Callable] | Unset:
        if isinstance(strats, Unset):
            return strats
        callables = []
        for strat in strats:
            if not strat.endswith(")"):
                strat += "()"  # noqa: PLW2901
            # TODO: Parse AST instead of using eval.
            callables.append(eval(f"Backlog.SortStrategy.{strat}"))  # noqa: S307
        return callables

    @classmethod
    def _get(cls, data: dict, *keys: str, default: Unset, transform: Callable[[Any], Any] | None = None) -> Any:
        """Get a value from a nested dictionary."""
        for key in keys:
            if key not in data:
                return default
            data = data[key]
        if transform:
            return transform(data)
        return data

    @classmethod
    def from_file(cls, path: str | Path) -> Config:
        """Load configuration from a file."""
        # Load data from the file.
        with open(path, "rb") as file:
            data = tomllib.load(file)

        # Check for unknown configuration keys.
        field_keys = [field.default.key for field in fields(cls)]  # type: ignore[union-attr]
        unknown_keys = []
        for top_level_key, top_level_value in data.items():
            for key in top_level_value.keys():  # noqa: SIM118
                final_key = f"{top_level_key}.{key}"
                if final_key not in field_keys:
                    unknown_keys.append(final_key)
        if unknown_keys:
            logger.warning(f"Unknown configuration keys: {', '.join(unknown_keys)}")

        # Create a configuration instance.
        return cls(
            **{
                field.name: cls._get(
                    data,
                    *field.default.key.split("."),  # type: ignore[union-attr]
                    default=field.default,  # type: ignore[arg-type]
                    transform=getattr(cls, field.default.transform or "", None),  # type: ignore[union-attr]
                )
                for field in fields(cls)
            },
        )

    @classmethod
    def from_default_location(cls) -> Config:
        """Load configuration from the default location."""
        if DEFAULT_CONF_PATH.exists():
            return cls.from_file(DEFAULT_CONF_PATH)
        return cls()



from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING
from typing import Annotated as An

import httpx
from typing_extensions import Doc

from insiders._internal.clients import Client
from insiders._internal.logger import logger
from insiders._internal.models import Issue, Sponsors, Sponsorship, User

if TYPE_CHECKING:
    from collections.abc import Iterable, Mapping

    from insiders._internal.models import IssueDict


class Polar(Client):
    """Polar client."""

    name = "Polar"

    def __init__(
        self,
        token: An[str, Doc("A Polar API token. Recommended scopes: `user:read`, `issues:read`, `subscriptions:read`.")],
    ) -> None:
        """Initialize Polar API client."""
        self.http_client = httpx.Client(
            base_url="https://api.polar.sh",
            headers={
                "Accept": "application/json",
                "Authorization": f"Bearer {token}",
            },
        )

    def get_sponsors(
        self,
        org_members_map: Mapping[str, Iterable[str]] | None = None,  # noqa: ARG002
        *,
        exclude_private: bool = False,  # noqa: ARG002
    ) -> An[Sponsors, Doc("Sponsors data.")]:
        """Get Polar sponsorships."""
        logger.debug("Fetching sponsors from Polar.")
        sponsorships = []
        page = 1
        items = []

        while True:
            logger.debug(f"Fetching page {page} of subscriptions from Polar.")
            response = self.http_client.get(
                "/v1/subscriptions/",
                params={
                    "active": True,
                    "sorting": "-started_at",  # To maintain order across pages.
                    "limit": 100,
                    "page": page,
                },
            )
            response.raise_for_status()
            data = response.json()
            items.extend(data["items"])
            if len(data["items"]) < 100:  # noqa: PLR2004
                break
            page += 1

        logger.debug(f"Processing {len(items)} subscriptions from Polar.")
        # Process sponsors data.
        for item in items:
            if not item["price"].get("price_amount"):
                continue

            # Determine account.
            account = User(
                name=item["user"]["public_name"],
                image=item["user"]["avatar_url"],
                url=f"https://polar.sh/{item['user']['public_name']}",
                platform="polar",
                platform_names={
                    "github": item["user"]["github_username"],
                },
            )
            logger.debug(f"Found user: @{account.name}")

            # Record sponsorship.
            sponsorships.append(
                Sponsorship(
                    private=False,
                    created=datetime.strptime(item["created_at"], "%Y-%m-%dT%H:%M:%S.%fZ"),  # noqa: DTZ007
                    amount=int(item["price"]["price_amount"] / 100),  # Polar stores in cents.
                    account=account,
                ),
            )

        return Sponsors(sponsorships=sponsorships)

    # Polar only supports fetching issues from GitHub for now.
    # Later, if it supports other platforms, we'll need to update this method.
    def get_issues(
        self,
        github_accounts: An[Iterable[str], Doc("GitHub accounts to fetch issues from.")],
        known_github_users: An[Iterable[User] | None, Doc("Known GitHub user accounts.")] = None,
    ) -> An[IssueDict, Doc("Issues data.")]:
        """Get issues from Polar."""
        page = 1
        items = []
        while True:
            logger.debug(f"Fetching page {page} of issues from Polar.")
            response = self.http_client.get(
                "/v1/issues/",
                params={
                    "external_organization_name": list(github_accounts),
                    "is_badged": True,
                    "sorting": "-created_at",  # To maintain order across pages.
                    "limit": 100,
                    "page": page,
                },
            )
            response.raise_for_status()
            data = response.json()
            items.extend(data["items"])
            if len(data["items"]) < 100:  # noqa: PLR2004
                break
            page += 1

        known_users = {account.name: account for account in (known_github_users or ())}
        issues = {}

        logger.debug(f"Processing {len(items)} issues from Polar.")
        for item in items:
            if item["state"] != "open":
                continue
            account_id = item["author"]["login"].removesuffix("[bot]")
            if account_id not in known_users:
                known_users[account_id] = User(name=account_id, platform="github")
            account = known_users[account_id]

            repository_name = f"{item['repository']['organization']['name']}/{item['repository']['name']}"
            iid = (repository_name, item["number"])
            issues[iid] = Issue(
                repository=repository_name,
                number=item["number"],
                title=item["title"],
                author=account,
                created=datetime.strptime(item["issue_created_at"], "%Y-%m-%dT%H:%M:%SZ"),  # noqa: DTZ007
                pledged=int(item["funding"]["pledges_sum"]["amount"] / 100),  # Polar stores in cents.
                platform="polar",
            )

        return issues



"""Sponsors management."""

from __future__ import annotations

from collections import defaultdict
from dataclasses import dataclass, field
from fnmatch import fnmatch
from typing import TYPE_CHECKING, Any, Callable, Literal
from typing import Annotated as An

from typing_extensions import Doc

if TYPE_CHECKING:
    from datetime import datetime


SponsorshipPlatform = An[Literal["github", "polar", "kofi"], Doc("The supported sponsorship platforms.")]
IssuePlatform = An[Literal["github", "polar"], Doc("The supported issue platforms.")]


@dataclass(kw_only=True, eq=False, frozen=True)
class Account:
    """An account."""

    name: An[str, Doc("The name of the account.")]
    image: An[str | None, Doc("The image URL of the account.")] = None
    url: An[str | None, Doc("The URL of the account.")] = None
    platform: An[SponsorshipPlatform, Doc("The platform of the account.")]
    sponsorships: An[
        list[Sponsorship],
        Doc("List of sponsorships associated with the account"),
    ] = field(default_factory=list)
    platform_names: An[
        dict[SponsorshipPlatform, str],
        Doc("Names of the account on different platforms."),
    ] = field(default_factory=dict)

    def __post_init__(self):
        self.platform_names[self.platform] = self.name

    def __eq__(self, other: An[object, Doc("The other object to compare to.")]) -> bool:
        if not isinstance(other, Account):
            return NotImplemented
        return self.platform == other.platform and self.name == other.name

    def __hash__(self):
        return hash((self.platform, self.name))

    @property
    def is_org(self) -> bool:
        """Return whether the account is an org."""
        return isinstance(self, Org)

    @property
    def is_user(self) -> bool:
        """Return whether the account is a user."""
        return isinstance(self, User)

    @property
    def direct_sponsor(self) -> bool:
        """Return whether the account is a direct sponsor."""
        return any(sponsorship.account is self for sponsorship in self.sponsorships)

    @property
    def highest_tier(self) -> int:
        """Return the highest tier amount."""
        return max((sponsorship.amount for sponsorship in self.sponsorships), default=0)

    @property
    def tier_sum(self) -> int:
        """Return the sum of all tier amounts."""
        return sum((sponsorship.amount for sponsorship in self.sponsorships), start=0)


@dataclass(kw_only=True, eq=False, frozen=True)
class User(Account):
    """A user account."""


@dataclass(kw_only=True, eq=False, frozen=True)
class Org(Account):
    """An org account."""

    users: An[dict[str, tuple[User, bool]], Doc("Users of this organization.")] = field(default_factory=dict)


@dataclass(kw_only=True, eq=False, frozen=True)
class Sponsorship:
    """A sponsorship."""

    private: An[bool, Doc("Indicates if the sponsorship is private")] = True
    created: An[datetime, Doc("The creation date of the sponsorship")]
    amount: An[int, Doc("The amount of the sponsorship")]
    account: An[User | Org, Doc("The user or org account associated with the sponsorship")]
    linked_accounts: An[dict[str, User | Org], Doc("Linked user accounts")] = field(default_factory=dict)

    def __eq__(self, other: An[object, Doc("The other object to compare to.")]) -> bool:
        if not isinstance(other, Sponsorship):
            return NotImplemented
        return self.account == other.account

    def __hash__(self):
        return hash(self.account)

    @property
    def users(self) -> set[User]:
        """Set of user accounts affected by this sponsorship."""
        users = set()
        for account in (self.account, *self.linked_accounts.values()):
            if isinstance(account, User):
                users.add(account)
            else:
                for user, _ in account.users.values():
                    users.add(user)
        return users


@dataclass(kw_only=True, eq=False, frozen=True)
class Sponsors:
    """Wrapper class for sponsorships."""

    sponsorships: An[list[Sponsorship], Doc("Sponsorships.")] = field(default_factory=list)

    def __post_init__(self) -> None:
        self.crosslink(create_missing_users=False)

    def merge(self, other: Sponsors, *, crosslink: bool = True) -> Sponsors:
        self.sponsorships.extend(other.sponsorships)
        if crosslink:
            self.crosslink(create_missing_users=True)
        return self

    @property
    def users(self) -> set[User]:
        """Set of user accounts affected by all sponsorships."""
        return {user for sponsorship in self.sponsorships for user in sponsorship.users}

    def crosslink(self, *, create_missing_users: bool = False) -> None:
        users_by_platform: dict[str, dict[str, User]] = defaultdict(dict)
        for user in self.users:
            users_by_platform[user.platform][user.name] = user

        for sponsorship in self.sponsorships:
            for platform, name in sponsorship.account.platform_names.items():
                if platform != sponsorship.account.platform:
                    if name not in users_by_platform[platform]:
                        if not create_missing_users:
                            continue
                        # Create missing user.
                        users_by_platform[platform][name] = User(name=name, platform=platform)
                    user = users_by_platform[platform][name]
                    # Link sponsorship to user.
                    sponsorship.linked_accounts[platform] = user
                    # Link user to sponsorship.
                    if sponsorship not in user.sponsorships:
                        user.sponsorships.append(sponsorship)


@dataclass(kw_only=True, eq=False)
class Issue:
    """An issue."""

    repository: An[str, Doc("The issue repository.")]
    number: An[int, Doc("The issue number.")]
    title: An[str, Doc("The issue title.")]
    created: An[datetime, Doc("The issue creation date.")]
    author: An[User, Doc("The issue author.")]
    upvotes: An[set[User], Doc("The issue upvotes / upvoters.")] = field(default_factory=set)
    labels: An[set[str], Doc("The issue labels.")] = field(default_factory=set)
    pledged: An[int, Doc("The amount pledged.")] = 0
    platform: An[IssuePlatform, Doc("The issue platform.")]

    @property
    def interested_users(self) -> set[User]:
        return {self.author, *self.upvotes}

    @property
    def sponsorships(self) -> set[Sponsorship]:
        return {sponsorship for user in self.interested_users for sponsorship in user.sponsorships}

    @property
    def funding(self) -> int:
        return sum(sponsorship.amount for sponsorship in self.sponsorships)


IssueDict = An[dict[tuple[str, str], Issue], Doc("A dictionary of issues.")]


@dataclass(kw_only=True, eq=False, frozen=True)
class Backlog:
    """Backlog of issues."""

    issues: An[list[Issue], Doc("A list of issues.")] = field(default_factory=list)

    class SortStrategy:
        @staticmethod
        def min_author_sponsorships(amount: int, *, reverse: bool = True) -> Callable[[Issue], int]:
            factor = -1 if reverse else 1
            return lambda issue: factor * (issue.author.tier_sum if issue.author.tier_sum >= amount else 0)

        @staticmethod
        def author_sponsorships(*, reverse: bool = True) -> Callable[[Issue], int]:
            factor = -1 if reverse else 1
            return lambda issue: factor * issue.author.tier_sum

        @staticmethod
        def min_upvoters_sponsorships(amount: int, *, reverse: bool = True) -> Callable[[Issue], int]:
            factor = -1 if reverse else 1
            return lambda issue: factor * (
                total if (total := sum(upvoter.tier_sum for upvoter in issue.upvotes)) >= amount else 0
            )

        @staticmethod
        def upvoters_sponsorships(*, reverse: bool = True) -> Callable[[Issue], int]:
            factor = -1 if reverse else 1
            return lambda issue: factor * sum(upvoter.tier_sum for upvoter in issue.upvotes)

        @staticmethod
        def min_sponsorships(amount: int, *, reverse: bool = True) -> Callable[[Issue], int]:
            factor = -1 if reverse else 1
            return lambda issue: factor * (total if (total := issue.funding) >= amount else 0)

        @staticmethod
        def sponsorships(*, reverse: bool = True) -> Callable[[Issue], int]:
            factor = -1 if reverse else 1
            return lambda issue: factor * issue.funding

        @staticmethod
        def min_pledge(amount: int, *, reverse: bool = True) -> Callable[[Issue], int]:
            factor = -1 if reverse else 1
            return lambda issue: factor * (issue.pledged if issue.pledged >= amount else 0)

        @staticmethod
        def pledge(*, reverse: bool = True) -> Callable[[Issue], int]:
            factor = -1 if reverse else 1
            return lambda issue: factor * issue.pledged

        @staticmethod
        def min_upvotes(amount: int, *, reverse: bool = True) -> Callable[[Issue], int]:
            factor = -1 if reverse else 1
            return lambda issue: factor * (upvotes if (upvotes := len(issue.upvotes)) >= amount else 0)

        @staticmethod
        def upvotes(*, reverse: bool = True) -> Callable[[Issue], int]:
            factor = -1 if reverse else 1
            return lambda issue: factor * len(issue.upvotes)

        @staticmethod
        def created(*, reverse: bool = False) -> Callable[[Issue], int]:
            factor = -1 if reverse else 1
            return lambda issue: factor * int(issue.created.timestamp())

        @staticmethod
        def label(name: str, *, reverse: bool = True) -> Callable[[Issue], int]:
            factor = -1 if reverse else 1
            return lambda issue: factor * (1 if name in issue.labels else 0)

        @staticmethod
        def repository(name: str, *, reverse: bool = True) -> Callable[[Issue], int]:
            factor = -1 if reverse else 1
            return lambda issue: factor * (1 if fnmatch(issue.repository, name) else 0)

    class _Sort:
        def __init__(self, *funcs: Callable[[Issue], Any]):
            self.funcs = list(funcs)

        def __call__(self, issue: Issue) -> tuple:
            return tuple(func(issue) for func in self.funcs)

    def sort(self, *strats: Callable[[Issue], Any]) -> None:
        """Sort the backlog."""
        self.issues.sort(key=self._Sort(*strats))



"""Module that contains the command line application."""

# Why does this file exist, and why not put this in `__main__`?
#
# You might be tempted to import things from `__main__` later,
# but that will cause problems: the code will get executed twice:
#
# - When you run `python -m insiders` python will execute
#   `__main__.py` as a script. That means there won't be any
#   `insiders.__main__` in `sys.modules`.
# - When you import `__main__` it will get executed again (as a module) because
#   there's no `insiders.__main__` in `sys.modules`.

from __future__ import annotations

import json
import shlex
import sys
from contextlib import nullcontext
from dataclasses import dataclass, field
from functools import wraps
from inspect import cleandoc
from pathlib import Path  # noqa: TC003
from typing import Annotated as An
from typing import Any, Callable, ClassVar, Literal

import cappa
from rich.console import Console
from typing_extensions import Doc

from insiders._internal import debug, defaults
from insiders._internal.clients import pypi
from insiders._internal.clients.github import GitHub
from insiders._internal.clients.index import Index
from insiders._internal.clients.polar import Polar
from insiders._internal.config import Config, Unset
from insiders._internal.logger import configure_logging

# TODO: Use PEP 727 everywhere.
# TODO: Re-organize all this.
from insiders._internal.ops.backlog import get_backlog, print_backlog
from insiders._internal.ops.projects import new_public_and_insiders_github_projects

_GROUP_ARGUMENTS = (10, "Arguments")
_GROUP_OPTIONS = (20, "Options")
_GROUP_GLOBAL_OPTIONS = (30, "Global options")
_GROUP_SUBCOMMANDS = (40, "Subcommands")


@dataclass(frozen=True)
class FromConfig(cappa.ValueFrom):
    def __init__(self, field: Unset | property, /) -> None:
        attr_name = field.fget.__name__ if isinstance(field, property) else field.name  # type: ignore[union-attr]
        super().__init__(self._from_config, attr_name=attr_name)

    @staticmethod
    def _from_config(attr_name: str) -> Any:
        config = CommandMain._load_config()
        value = getattr(config, attr_name)
        return cappa.Empty if isinstance(value, Unset) else value


# ============================================================================ #
# Projects                                                                     #
# ============================================================================ #
@cappa.command(name="project", help="Manage projects (GitHub and local copies).")
@dataclass(kw_only=True)
class CommandProject:
    """Command to manage projects on GitHub and locally."""

    subcommand: An[
        CommandProjectCreate | CommandProjectCheck,
        cappa.Subcommand(group=_GROUP_SUBCOMMANDS),
        Doc("The selected subcommand."),
    ]


@cappa.command(
    name="create",
    help="Create public/insiders repositories.",
    description=cleandoc(
        """
        This command will do several things:

        - Create public and insiders repositories on GitHub
            (using the provided namespace, username, repository name, description, etc.).
        - Clone these two repositories locally (using the provided repository paths).
        - Optionally initialize the public repository by generating initial contents
            using the specified [Copier](https://copier.readthedocs.io/en/stable/) template and answers.
        - Optionally run a post creation command into the public repository.
        - Pull the public contents into the insiders clone (by declaring an `upstream` remote).

        *Example 1 - Project in user's namespace*

        The insiders namespace, insiders repository name and username are inferred
        from the namespace and repository name.

        ```bash
        insiders create \\
            -n pawamoy \\
            -r mkdocs-ultimate \\
            -d "The ultimate plugin for MkDocs (??)" \\
            -o ~/data/dev \\
            -O ~/data/dev/insiders \\
            -t gh:pawamoy/copier-uv
        ```

        *Example 2 - Project in another namespace:*

        The insiders namespace, insiders repository name and username are different,
        so must be provided explicitly:

        ```bash
        insiders create \\
            -n mkdocstrings \\
            -r rust \\
            -d "A Rust handler for mkdocstrings" \\
            -o ~/data/dev \\
            -O ~/data/dev/insiders \\
            -N pawamoy-insiders \\
            -R mkdocstrings-rust \\
            -u pawamoy \\
            -t gh:mkdocstrings/handler-template
        ```
        """,
    ),
)
@dataclass(kw_only=True)
class CommandProjectCreate:
    """Command to create public/insiders repositories."""

    repository: An[
        str,
        cappa.Arg(short="-r", long=True, group=_GROUP_OPTIONS),
        Doc("""Name of the public repository."""),
    ]

    description: An[
        str,
        cappa.Arg(short="-d", long=True, group=_GROUP_OPTIONS),
        Doc("""Shared description."""),
    ]

    namespace: An[
        str,
        cappa.Arg(
            short="-n",
            long=True,
            default=FromConfig(Config.github_project_namespace),
            show_default=f"{Config.github_project_namespace}",
            group=_GROUP_OPTIONS,
        ),
        Doc("""Namespace of the public repository."""),
    ]

    project_directory: An[
        Path,
        cappa.Arg(
            short="-o",
            long=True,
            default=FromConfig(Config.project_directory),
            show_default=f"{Config.project_directory}",
            group=_GROUP_OPTIONS,
        ),
        Doc("""Directory in which to clone the public repository."""),
    ]

    insiders_repository: An[
        str | None,
        cappa.Arg(short="-R", long=True, show_default="public name", group=_GROUP_OPTIONS),
        Doc("""Name of the insiders repository."""),
    ] = None

    insiders_namespace: An[
        str | None,
        cappa.Arg(
            short="-N",
            long=True,
            default=FromConfig(Config.github_insiders_project_namespace),
            show_default=f"{Config.github_insiders_project_namespace} or public namespace",
            group=_GROUP_OPTIONS,
        ),
        Doc("""Namespace of the insiders repository."""),
    ] = None

    insiders_project_directory: An[  # type: ignore[misc]
        Path,
        cappa.Arg(
            short="-O",
            long=True,
            default=FromConfig(Config.project_insiders_directory),
            show_default=f"{Config.project_insiders_directory}",
            group=_GROUP_OPTIONS,
        ),
        Doc("""Directory in which to clone the insiders repository."""),
    ]

    github_username: An[
        str | None,
        cappa.Arg(
            short="-u",
            long=True,
            default=FromConfig(Config.github_username),
            show_default=f"{Config.github_username} or public namespace",
            group=_GROUP_OPTIONS,
        ),
        Doc("""GitHub username."""),
    ] = None

    copier_template: An[
        str | None,
        cappa.Arg(
            short="-t",
            long=True,
            default=FromConfig(Config.project_copier_template),
            show_default=f"{Config.project_copier_template}",
            group=_GROUP_OPTIONS,
        ),
        Doc("""Copier template to generate new projects with."""),
    ] = None

    @staticmethod
    def _parse_dict(arg: str) -> dict[str, str]:
        return dict(pair.split("=", 1) for pair in arg.split(","))

    copier_template_answers: An[
        dict[str, str] | None,
        cappa.Arg(
            short="-a",
            long=True,
            parse=_parse_dict,
            default=FromConfig(Config.project_copier_template_answers),
            show_default=f"{Config.project_copier_template_answers}",
            group=_GROUP_OPTIONS,
        ),
        Doc("""Copier template answers to use when generating a project."""),
    ] = None

    post_creation_command: An[
        list[str] | None,
        cappa.Arg(
            short="-x",
            long=True,
            num_args=1,
            parse=shlex.split,
            default=FromConfig(Config.project_post_creation_command),
            show_default=f"{Config.project_post_creation_command}",
            group=_GROUP_OPTIONS,
        ),
        Doc("""Command to run after creating the public repository."""),
    ] = None

    register_on_pypi: An[
        bool,
        cappa.Arg(
            short="-i",
            long=True,
            default=FromConfig(Config.project_register_on_pypi),
            show_default=f"{Config.project_register_on_pypi}",
            group=_GROUP_OPTIONS,
        ),
        Doc("""Whether to register the project on PyPI after creating it."""),
    ] = False

    pypi_username: An[
        str | None,
        cappa.Arg(
            short="-y",
            long=True,
            default=FromConfig(Config.pypi_username),
            show_default=f"{Config.pypi_username}",
            group=_GROUP_OPTIONS,
        ),
        Doc("""PyPI username to register the project with."""),
    ] = None

    def __call__(self) -> int:
        new_public_and_insiders_github_projects(
            public_namespace=self.namespace,
            public_name=self.repository,
            description=self.description,
            # We use the Insiders name here as it will generally be more independant of the namespace
            # (for example `mkdocstrings-python` instead of just `python`).
            public_repo_path=self.project_directory / (self.insiders_repository or self.repository),
            insiders_namespace=self.insiders_namespace,
            insiders_name=self.insiders_repository,
            insiders_repo_path=self.insiders_project_directory / (self.insiders_repository or self.repository),
            github_username=self.github_username,
            copier_template=self.copier_template,
            copier_template_answers=self.copier_template_answers,
            post_creation_command=self.post_creation_command,
        )
        if self.register_on_pypi:
            if not self.pypi_username:
                raise cappa.Exit("PyPI username must be provided to register the project on PyPI.", code=1)
            pypi.reserve_pypi(
                username=self.pypi_username,
                name=self.repository,
                description=self.description,
            )
        return 0


@cappa.command(
    name="check",
    help="Check public/insiders repositories.",
    description=cleandoc(
        """
        TODO. Check that everything is consistent.
        """,
    ),
)
@dataclass(kw_only=True)
class CommandProjectCheck:
    """Command to check GitHub projects."""

    def __call__(self) -> int:
        raise cappa.Exit("Not implemented yet.", code=1)


# ============================================================================ #
# PyPI                                                                         #
# ============================================================================ #
@cappa.command(name="pypi", help="Manage PyPI-related things.")
@dataclass(kw_only=True)
class CommandPyPI:
    """Command to manage PyPI-related things."""

    subcommand: An[
        CommandPyPIRegister,
        cappa.Subcommand(group=_GROUP_SUBCOMMANDS),
        Doc("The selected subcommand."),
    ]


@cappa.command(
    name="register",
    help="Register a name on PyPI.",
    description=cleandoc(
        """
        This will create a temporary project on your filesystem,
        then build both source and wheel distributions for it,
        and upload them to PyPI using Twine.

        After that, you will see an initial version 0.0.0
        of your project on PyPI.

        *Example*

        ```bash
        insiders pypi register -u pawamoy -n my-new-project -d "My new project!"
        ```

        Credentials must be configured in `~/.pypirc` to allow Twine to push to PyPI.
        For example, if you use [PyPI API tokens](https://pypi.org/help/#apitoken),
        add the token to your keyring:

        ```bash
        pipx install keyring
        keyring set https://upload.pypi.org/legacy/ __token__
        # __token__ is a literal string, do not replace it with your token.
        # The command will prompt you to paste your token.
        ```

        And configure `~/.pypirc`:

        ```ini
        [distutils]
        index-servers =
            pypi

        [pypi]
        username: __token__
        ```
        """,
    ),
)
@dataclass(kw_only=True)
class CommandPyPIRegister:
    """Command to register a project name on PyPI."""

    username: An[
        str,
        cappa.Arg(
            short="-u",
            long=True,
            default=FromConfig(Config.pypi_username),
            show_default=f"{Config.pypi_username}",
            group=_GROUP_OPTIONS,
        ),
        Doc("Username on PyPI (your account)."),
    ]

    name: An[
        str,
        cappa.Arg(short="-n", long=True, group=_GROUP_OPTIONS),
        Doc("Name to register."),
    ]

    description: An[
        str,
        cappa.Arg(short="-d", long=True, group=_GROUP_OPTIONS),
        Doc("Description of the project on PyPI."),
    ]

    def __call__(self) -> Any:
        pypi.reserve_pypi(self.username, self.name, self.description)
        return 0


# ============================================================================ #
# Index                                                                        #
# ============================================================================ #
@cappa.command(name="index", help="Manage the local index.")
@dataclass(kw_only=True)
class CommandIndex:
    """Command to manage the local index."""

    subcommand: An[
        CommandIndexList
        | CommandIndexAdd
        | CommandIndexRemove
        | CommandIndexUpdate
        | CommandIndexStart
        | CommandIndexStatus
        | CommandIndexStop
        | CommandIndexLogs,
        cappa.Subcommand(group=_GROUP_SUBCOMMANDS),
        Doc("The selected subcommand."),
    ]


@cappa.command(
    name="list",
    help="List insiders repositories.",
    description="List the watched repositories.",
)
@dataclass(kw_only=True)
class CommandIndexList:
    """Command to list the watched repositories."""

    sources_directory: An[
        Path,
        cappa.Arg(
            short="-s",
            long=True,
            default=FromConfig(Config.index_sources_directory),
            show_default=f"{Config.index_sources_directory} or `{{default}}`",
            group=_GROUP_OPTIONS,
        ),
        Doc("Directory where the sources are stored."),
    ] = defaults.DEFAULT_REPO_DIR

    distributions_directory: An[
        Path,
        cappa.Arg(
            short="-d",
            long=True,
            default=FromConfig(Config.index_distributions_directory),
            show_default=f"{Config.index_distributions_directory} or `{{default}}`",
            group=_GROUP_OPTIONS,
        ),
        Doc("Directory where the distributions are stored."),
    ] = defaults.DEFAULT_DIST_DIR

    dists: An[
        bool,
        cappa.Arg(short="-i", long=True, show_default="True", group=_GROUP_OPTIONS),
        Doc("List distributions."),
    ] = False

    projects: An[
        bool,
        cappa.Arg(short="-p", long=True, show_default="True", group=_GROUP_OPTIONS),
        Doc("List projects."),
    ] = False

    def __call__(self) -> int:
        index = Index(git_dir=self.sources_directory, dist_dir=self.distributions_directory)
        if self.dists is self.projects:
            print("Distributions:")
            for dist in sorted(index.list_distributions()):
                print(dist)
            print("\nProjects:")
            for project in sorted(index.list_projects()):
                print(project)
        elif self.dists:
            for dist in sorted(index.list_distributions()):
                print(dist)
        elif self.projects:
            for project in sorted(index.list_projects()):
                print(project)
        return 0


@cappa.command(
    name="add",
    help="Add insiders repositories.",
    description="Add a repository to the watched repositories.",
)
@dataclass(kw_only=True)
class CommandIndexAdd:
    """Command to add a repository to the watched repositories."""

    repositories: An[
        list[str],
        cappa.Arg(group=_GROUP_ARGUMENTS),
        Doc("List of repositories (GitHub namespace/project or Git URL git@host:repo)."),
    ]

    sources_directory: An[
        Path,
        cappa.Arg(
            short="-s",
            long=True,
            default=FromConfig(Config.index_sources_directory),
            show_default=f"{Config.index_sources_directory} or `{{default}}`",
            group=_GROUP_OPTIONS,
        ),
        Doc("Directory where the sources are stored."),
    ] = defaults.DEFAULT_REPO_DIR

    distributions_directory: An[
        Path,
        cappa.Arg(
            short="-d",
            long=True,
            default=FromConfig(Config.index_distributions_directory),
            show_default=f"{Config.index_distributions_directory} or `{{default}}`",
            group=_GROUP_OPTIONS,
        ),
        Doc("Directory where the distributions are stored."),
    ] = defaults.DEFAULT_DIST_DIR

    url: An[
        str,
        cappa.Arg(
            short="-u",
            long=True,
            default=FromConfig(Config.index_url),
            show_default=f"{Config.index_url} or `{{default}}`",
            group=_GROUP_OPTIONS,
        ),
        Doc("URL of the index to upload packages to."),
    ] = defaults.DEFAULT_INDEX_URL

    def __call__(self) -> int:
        index = Index(url=self.url, git_dir=self.sources_directory, dist_dir=self.distributions_directory)
        for project in self.repositories:
            index.add(project if project.startswith("git@") else f"git@github.com:{project}")
        return 0


@cappa.command(
    name="remove",
    help="Remove insiders repositories.",
    description="Remove a repository from the watched repositories.",
)
@dataclass(kw_only=True)
class CommandIndexRemove:
    """Command to remove a repository and its distributions (if served locally)."""

    repositories: An[
        list[str],
        cappa.Arg(group=_GROUP_ARGUMENTS),
        Doc("List of repository names."),
    ]

    sources_directory: An[
        Path,
        cappa.Arg(
            short="-s",
            long=True,
            default=FromConfig(Config.index_sources_directory),
            show_default=f"{Config.index_sources_directory} or `{{default}}`",
            group=_GROUP_OPTIONS,
        ),
        Doc("Directory where the sources are stored."),
    ] = defaults.DEFAULT_REPO_DIR

    distributions_directory: An[
        Path,
        cappa.Arg(
            short="-d",
            long=True,
            default=FromConfig(Config.index_distributions_directory),
            show_default=f"{Config.index_distributions_directory} or `{{default}}`",
            group=_GROUP_OPTIONS,
        ),
        Doc("Directory where the distributions are stored."),
    ] = defaults.DEFAULT_DIST_DIR

    def __call__(self) -> int:
        index = Index(git_dir=self.sources_directory, dist_dir=self.distributions_directory)
        for repo in self.repositories:
            index.remove(repo)
        return 0


@cappa.command(
    name="update",
    help="Update insiders packages.",
    description="Update watched projects.",
)
@dataclass(kw_only=True)
class CommandIndexUpdate:
    """Command to update watched projects."""

    repositories: An[
        list[str],
        cappa.Arg(group=_GROUP_ARGUMENTS),
        Doc("List of repository names."),
    ] = field(default_factory=list)

    sources_directory: An[
        Path,
        cappa.Arg(
            short="-s",
            long=True,
            default=FromConfig(Config.index_sources_directory),
            show_default=f"{Config.index_sources_directory} or `{{default}}`",
            group=_GROUP_OPTIONS,
        ),
        Doc("Directory where the sources are stored."),
    ] = defaults.DEFAULT_REPO_DIR

    distributions_directory: An[
        Path,
        cappa.Arg(
            short="-d",
            long=True,
            default=FromConfig(Config.index_distributions_directory),
            show_default=f"{Config.index_distributions_directory} or `{{default}}`",
            group=_GROUP_OPTIONS,
        ),
        Doc("Directory where the distributions are stored."),
    ] = defaults.DEFAULT_DIST_DIR

    url: An[
        str,
        cappa.Arg(
            short="-u",
            long=True,
            default=FromConfig(Config.index_url),
            show_default=f"{Config.index_url} or `{{default}}`",
            group=_GROUP_OPTIONS,
        ),
        Doc("URL of the index to upload packages to."),
    ] = defaults.DEFAULT_INDEX_URL

    def __call__(self) -> int:
        index = Index(url=self.url, git_dir=self.sources_directory, dist_dir=self.distributions_directory)
        index.update(self.repositories)
        return 0


@cappa.command(
    name="start",
    help="Start the server.",
    description="Start the server in the background.",
)
@dataclass(kw_only=True)
class CommandIndexStart:
    """Command to start the server."""

    sources_directory: An[
        Path,
        cappa.Arg(
            short="-s",
            long=True,
            default=FromConfig(Config.index_sources_directory),
            show_default=f"{Config.index_sources_directory} or `{{default}}`",
            group=_GROUP_OPTIONS,
        ),
        Doc("Directory where the sources are stored."),
    ] = defaults.DEFAULT_REPO_DIR

    distributions_directory: An[
        Path,
        cappa.Arg(
            short="-d",
            long=True,
            default=FromConfig(Config.index_distributions_directory),
            show_default=f"{Config.index_distributions_directory} or `{{default}}`",
            group=_GROUP_OPTIONS,
        ),
        Doc("Directory where the distributions are stored."),
    ] = defaults.DEFAULT_DIST_DIR

    url: An[
        str,
        cappa.Arg(
            short="-u",
            long=True,
            default=FromConfig(Config.index_url),
            show_default=f"{Config.index_url} or `{{default}}`",
            group=_GROUP_OPTIONS,
        ),
        Doc("URL to serve the index at."),
    ] = defaults.DEFAULT_INDEX_URL

    background: An[
        bool,
        cappa.Arg(
            short="-b",
            long=True,
            default=FromConfig(Config.index_start_in_background),
            show_default=f"{Config.index_start_in_background} or {{default}}",
            group=_GROUP_OPTIONS,
        ),
        Doc("Run the server in the background."),
    ] = False

    log_path: An[
        str | None,
        cappa.Arg(
            short="-l",
            long=True,
            default=FromConfig(Config.index_log_path),
            show_default=f"{Config.index_log_path} or standard error",
            group=_GROUP_OPTIONS,
        ),
        Doc("Where to write index server logs."),
    ] = None

    def __call__(self) -> int:
        index = Index(url=self.url, git_dir=self.sources_directory, dist_dir=self.distributions_directory)
        index.start(background=self.background, log_path=self.log_path)
        return 0


@cappa.command(
    name="status",
    help="Show the server status.",
    description="Show the server status.",
)
@dataclass(kw_only=True)
class CommandIndexStatus:
    """Command to show the server status."""

    def __call__(self) -> int:
        proc_data = Index().status()
        if proc_data:
            print("Running:")
            print(json.dumps(proc_data, indent=2, sort_keys=True))
        else:
            print("Not running")
        return 0


@cappa.command(
    name="stop",
    help="Stop the server.",
    description="Stop the server.",
)
@dataclass(kw_only=True)
class CommandIndexStop:
    """Command to stop the server."""

    def __call__(self) -> int:
        return 0 if Index().stop() else 1


@cappa.command(
    name="logs",
    help="Show the server logs.",
    description="Show the server logs.",
)
@dataclass(kw_only=True)
class CommandIndexLogs:
    """Command to show the server logs."""

    def __call__(self) -> int:
        index = Index()
        try:
            print(index.logs())
        except FileNotFoundError as error:
            config = CommandMain._load_config()
            if isinstance(config.index_log_path, Unset):
                print(error, file=sys.stderr)
                return 1
            print(config.index_log_path)
        return 0


# ============================================================================ #
# Teams                                                                        #
# ============================================================================ #
@cappa.command(name="team", help="Manage GitHub teams.")
@dataclass(kw_only=True)
class CommandTeam:
    """Command to manage GitHub teams."""

    subcommand: An[cappa.Subcommands[CommandTeamList | CommandTeamSync], Doc("The selected subcommand.")]


@cappa.command(
    name="list",
    help="List members of a team.",
    description=cleandoc(
        """
        List the members of a GitHub team.
        """,
    ),
)
@dataclass(kw_only=True)
class CommandTeamList:
    """Command to list team memberships."""

    def __call__(self) -> int:
        raise cappa.Exit("Not implemented yet.", code=1)


@cappa.command(
    name="sync",
    help="Synchronize members of a team with current sponsors.",
    description=cleandoc(
        """
        Fetch current sponsors from GitHub,
        then grant or revoke access to a GitHub team
        for eligible sponsors.
        """,
    ),
)
@dataclass(kw_only=True)
class CommandTeamSync:
    """Command to sync team memberships with current sponsors."""

    github_insiders_team: An[
        str,
        cappa.Arg(
            short=False,
            long=False,
            num_args=1,
            default=FromConfig(Config.github_insiders_team),
            show_default=f"{Config.github_insiders_team}",
            group=_GROUP_ARGUMENTS,
        ),
        Doc("""The GitHub team to sync."""),
    ]

    github_sponsored_account: An[
        str,
        cappa.Arg(
            short=False,
            long=("--ghsa", "--github-sponsored-account"),
            default=FromConfig(Config.github_sponsored_account),
            show_default=f"{Config.github_sponsored_account} or none",
            group=_GROUP_OPTIONS,
        ),
        Doc("""The sponsored account on GitHub Sponsors."""),
    ] = ""

    github_include_users: An[
        list[str],
        cappa.Arg(
            short=False,
            long=("--ghiu", "--github-include-users"),
            default=FromConfig(Config.github_include_users),
            show_default=f"{Config.github_include_users}",
            group=_GROUP_OPTIONS,
        ),
        Doc("""Users that should always be in the team."""),
    ] = field(default_factory=list)

    github_exclude_users: An[
        list[str],
        cappa.Arg(
            short=False,
            long=("--gheu", "--github-exclude-users"),
            default=FromConfig(Config.github_exclude_users),
            show_default=f"{Config.github_exclude_users}",
            group=_GROUP_OPTIONS,
        ),
        Doc("""Users that should never be in the team."""),
    ] = field(default_factory=list)

    github_organization_members: An[
        dict[str, list[str]],
        cappa.Arg(
            short=False,
            long=("--ghom", "--github-organization-members"),
            default=FromConfig(Config.github_organization_members),
            show_default=f"{Config.github_organization_members}",
            group=_GROUP_OPTIONS,
        ),
        Doc("""A mapping of users belonging to sponsoring organizations."""),
    ] = field(default_factory=dict)

    github_token: An[
        str,
        cappa.Arg(
            short=False,
            long=("--ght", "--github-token"),
            default=cappa.Env("GITHUB_TOKEN") | FromConfig(Config.github_token),
            show_default="`GITHUB_TOKEN` env-var or `github.token-command` config-value",
            group=_GROUP_OPTIONS,
        ),
        Doc("""A GitHub token. Recommended scopes: `admin:org` and `read:user`."""),
    ] = ""

    polar_sponsored_account: An[
        str,
        cappa.Arg(
            short=False,
            long=("--plsa", "--polar-sponsored-account"),
            default=FromConfig(Config.polar_sponsored_account),
            show_default=f"{Config.polar_sponsored_account} or none",
            group=_GROUP_OPTIONS,
        ),
        Doc("""The sponsored account on Polar."""),
    ] = ""

    polar_token: An[
        str,
        cappa.Arg(
            short=False,
            long=("--plt", "--polar-token"),
            default=cappa.Env("POLAR_TOKEN") | FromConfig(Config.polar_token),
            show_default="`POLAR_TOKEN` env-var or `polar.token-command` config-value",
            group=_GROUP_OPTIONS,
        ),
        Doc("""A Polar token. Recommended scopes: `user:read`, `issues:read`, `subscriptions:read`."""),
    ] = ""

    minimum_amount: An[
        int,
        cappa.Arg(
            short=True,
            long=True,
            default=FromConfig(Config.sponsors_minimum_amount),
            show_default=f"{Config.sponsors_minimum_amount} or `{{default}}`",
            group=_GROUP_OPTIONS,
        ),
        Doc("""Minimum amount to be considered an insider."""),
    ] = 0

    def __call__(self) -> int:
        # TODO: Gather sponsors from configured platforms.
        with GitHub(self.github_token) as github:
            github.sync_team(
                self.github_insiders_team,
                min_amount=self.minimum_amount,
                include_users=set(self.github_include_users),
                exclude_users=set(self.github_exclude_users),
                org_users=self.github_organization_members,  # type: ignore[arg-type]
            )
        return 0


# ============================================================================ #
# Backlog                                                                      #
# ============================================================================ #
@cappa.command(
    name="backlog",
    help="List the backlog.",
    description=cleandoc(
        """
        List the issues in the backlog.
        """,
    ),
)
@dataclass(kw_only=True)
class CommandBacklog:
    """Command to list the backlog of issues."""

    @staticmethod
    def _parse_sort(arg: str) -> list[Callable]:
        return Config._eval_sort(arg.split(",")) or []

    backlog_namespaces: An[
        list[str],
        cappa.Arg(
            short=False,
            long=False,
            default=cappa.Env("BACKLOG_NAMESPACES") | FromConfig(Config.backlog_namespaces),
            show_default=f"`BACKLOG_NAMESPACES` env-var or {Config.backlog_namespaces}",
            group=_GROUP_ARGUMENTS,
        ),
        Doc("Namespaces to fetch issues from."),
    ]

    issue_labels: An[
        dict[str, str],
        cappa.Arg(
            short=True,
            long=True,
            default=FromConfig(Config.backlog_issue_labels),
            show_default=f"{Config.backlog_issue_labels}",
            group=_GROUP_OPTIONS,
        ),
        Doc("Issue labels to keep in issues metadata, and how they are represented."),
    ] = field(default_factory=dict)

    limit: An[
        int,
        cappa.Arg(
            short=True,
            long=True,
            default=FromConfig(Config.backlog_limit),
            show_default=f"{Config.backlog_limit} or `{{default}}`",
            group=_GROUP_OPTIONS,
        ),
        Doc("Limit the number of issues to display."),
    ] = 0

    sort: An[
        list[Callable],
        cappa.Arg(
            short=True,
            long=True,
            parse=_parse_sort,
            default=FromConfig(Config.backlog_sort),
            show_default=f"{Config.backlog_sort}",
            group=_GROUP_OPTIONS,
        ),
        Doc("Sort strategy."),
    ] = field(default_factory=list)

    public: An[
        bool,
        cappa.Arg(short=False, long=True, show_default=False, group=_GROUP_OPTIONS),
        Doc("Only use public sponsorships."),
    ] = False

    polar_token: An[
        str,
        cappa.Arg(
            short=False,
            long=("--plt", "--polar-token"),
            default=cappa.Env("POLAR_TOKEN") | FromConfig(Config.polar_token),
            show_default="`POLAR_TOKEN` env-var or `polar.token-command` config-value",
            group=_GROUP_OPTIONS,
        ),
        Doc("""A Polar token. Recommended scopes: `user:read`, `issues:read`, `subscriptions:read`."""),
    ] = ""

    github_token: An[
        str,
        cappa.Arg(
            short=False,
            long=("--ght", "--github-token"),
            default=cappa.Env("GITHUB_TOKEN") | FromConfig(Config.github_token),
            show_default="`GITHUB_TOKEN` env-var or `github.token-command` config-value",
            group=_GROUP_OPTIONS,
        ),
        Doc("""A GitHub token. Recommended scopes: `read:user`."""),
    ] = ""

    github_organization_members: An[
        dict[str, list[str]],
        cappa.Arg(
            short=False,
            long=("--ghom", "--github-organization-members"),
            default=FromConfig(Config.github_organization_members),
            show_default=f"{Config.github_organization_members}",
            group=_GROUP_OPTIONS,
        ),
        Doc("""A mapping of users belonging to sponsoring organizations."""),
    ] = field(default_factory=dict)

    def __call__(self) -> int:
        github_context = GitHub(self.github_token)
        polar_context = Polar(self.polar_token) if self.polar_token else nullcontext()
        with github_context as github, polar_context as polar, Console().status("") as status:
            status.update("Fetching sponsors from GitHub")
            sponsors = github.get_sponsors(self.github_organization_members, exclude_private=self.public)
            if polar:
                status.update("Fetching sponsors from Polar")
                sponsors.merge(polar.get_sponsors(exclude_private=self.public))
            status.update(f"Fetching issues from GitHub{' and Polar' if isinstance(polar, Polar) else ''}")
            backlog = get_backlog(
                self.backlog_namespaces,
                github=github,
                polar=polar,
                sponsors=sponsors,
                issue_labels=set(self.issue_labels),
            )
        if self.sort:
            status.update("Sorting issues")
            backlog.sort(*self.sort)
        print_backlog(backlog, self.issue_labels, limit=self.limit)
        return 0


# ============================================================================ #
# Main                                                                         #
# ============================================================================ #
@cappa.command(
    name="insiders",
    help="Manage your Insiders projects.",
    description=cleandoc(
        """
        This tool lets you manage your local and remote Git repositories
        for projects that offer an [Insiders](https://pawamoy.github.io/insiders/) version.

        See the documentation / help text of the different subcommands available.

        *Example*

        ```bash
        insiders --debug-info
        ```
        """,
    ),
)
@dataclass(kw_only=True)
class CommandMain:
    """Command to manage your Insiders projects."""

    _CONFIG: ClassVar[Config | None] = None

    @staticmethod
    def _load_config(file: Path | None = None) -> Config:
        if CommandMain._CONFIG is None:
            CommandMain._CONFIG = Config.from_file(file) if file else Config.from_default_location()
        return CommandMain._CONFIG

    subcommand: An[
        CommandProject | CommandPyPI | CommandIndex | CommandBacklog | CommandTeam,
        cappa.Subcommand(group=_GROUP_SUBCOMMANDS),
        Doc("The selected subcommand."),
    ]

    @staticmethod
    def _print_and_exit(func: Callable[[], str | None], code: int = 0) -> Callable[[], None]:
        @wraps(func)
        def _inner() -> None:
            raise cappa.Exit(func() or "", code=code)

        return _inner

    @staticmethod
    def _configure_logging(command: CommandMain) -> None:
        configure_logging(command.log_level, command.log_path)

    version: An[
        bool,
        cappa.Arg(
            short="-V",
            long=True,
            action=_print_and_exit(debug.get_version),
            num_args=0,
            group=_GROUP_OPTIONS,
        ),
        Doc("Print the program version and exit."),
    ] = False

    debug_info: An[
        bool,
        cappa.Arg(long=True, action=_print_and_exit(debug.print_debug_info), num_args=0, group=_GROUP_OPTIONS),
        Doc("Print debug information."),
    ] = False

    config: An[
        Config,
        cappa.Arg(
            short="-c",
            long=True,
            parse=_load_config,
            propagate=True,
            show_default=f"`{defaults.DEFAULT_CONF_PATH}`",
            group=_GROUP_GLOBAL_OPTIONS,
        ),
        Doc("Path to the configuration file."),
    ] = field(default_factory=Config.from_default_location)

    log_level: An[
        Literal["TRACE", "DEBUG", "INFO", "SUCCESS", "WARNING", "ERROR", "CRITICAL"],
        cappa.Arg(
            short="-L",
            long=True,
            parse=str.upper,
            propagate=True,
            show_default=True,
            group=_GROUP_GLOBAL_OPTIONS,
        ),
        Doc("Log level to use when logging messages."),
    ] = "INFO"

    log_path: An[
        str | None,
        cappa.Arg(short="-P", long=True, propagate=True, show_default="standard error", group=_GROUP_GLOBAL_OPTIONS),
        Doc("Write log messages to this file path."),
    ] = None


def main(
    args: An[list[str] | None, Doc("Arguments passed from the command line.")] = None,
) -> An[int, Doc("An exit code.")]:
    """Run the main program.

    This function is executed when you type `insiders` or `python -m insiders`.
    """
    output = cappa.Output(error_format="[bold]insiders[/]: [bold red]error[/]: {message}")
    completion_option: cappa.Arg = cappa.Arg(
        long=True,
        action=cappa.ArgAction.completion,
        choices=["complete", "generate"],
        group=_GROUP_GLOBAL_OPTIONS,
        help="Print shell-specific completion source.",
    )
    help_option: cappa.Arg = cappa.Arg(
        short="-h",
        long=True,
        action=cappa.ArgAction.help,
        group=_GROUP_GLOBAL_OPTIONS,
        help="Print the program help and exit.",
    )
    help_formatter = cappa.HelpFormatter(default_format="Default: {default}.")

    try:
        return cappa.invoke(
            CommandMain,
            argv=args,
            output=output,
            help=help_option,
            completion=completion_option,
            help_formatter=help_formatter,
            deps=[CommandMain._configure_logging],
        )
    except cappa.Exit as exit:
        return int(exit.code or 0)
